from math import floor

import numpy as np
from scipy.signal import convolve, convolve2d


def EnvFilt(xdB, ydB, fcut, fsub, fsamp):
    """
    Function to lowpass filter and subsample the envelope in dB SL produced
    by the model of the auditory periphery. The LP filter uses a von Hann
    raised cosine window to ensure that there are no negative envelope values
    produced by the filtering operation.

    Args:
        xdB (np.ndarray): env in dB SL for the ref signal in each auditory band
        ydB (np.ndarray): env in dB SL for the degraded signal in each auditory band
        fcut ():  LP filter cutoff frequency for the filtered envelope, Hz
        fsub ():  subsampling frequency in Hz for the LP filtered envelopes
        fsamp ():  sampling rate in Hz for the signals xdB and ydB

    Returns:
        tuple: xLP - LP filtered and subsampled reference signal envelope
           Each frequency band is a separate column. yLP - LP filtered and subsampled
           degraded signal envelope


    James M. Kates, 12 September 2019.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Check the filter design parameters
    assert fsub <= fsamp, "Error in ebm.EnvFilt: Subsampling rate too high."
    assert fcut <= 0.5 * fsub, "Error in ebm.EnvFilt: LP cutoff frequency too high."

    # Check the data matrix orientation
    # Require each frequency band to be a separate column
    nrow = xdB.shape[0]  # number of rows
    ncol = xdB.shape[1]  # number of columnts
    if ncol > nrow:
        xdB = xdB.T
        ydB = ydB.T
    nsamp = xdB.shape[0]

    # Compute the lowpass filter length in samples to give -3 dB at fcut Hz
    tfilt = 1000 * (1 / fcut)  # filter length in ms
    tfilt = 0.7 * tfilt  # Empirical adjustment to the filter length
    nfilt = round(0.001 * tfilt * fsamp)  # Filter length in samples
    nhalf = floor(nfilt / 2)
    nfilt = int(2 * nhalf)  # Force an even filter length

    # Design the FIR LP filter using a von Hann window to ensure that there are no negative
    # envelope values. The MATLAB code uses the hanning() function, which returns the Hann
    # window without the first and last zero-weighted window samples,
    # unlike np.hann and scipy.signal.windows.hann; the code below replicates this behaviour
    w = 0.5 * (1 - np.cos(2 * np.pi * np.arange(1, nfilt / 2 + 1) / (nfilt + 1)))
    benv = np.concatenate((w, np.flip(w)))
    benv = benv / np.sum(benv)

    # LP filter for the envelopes at fsamp
    xenv = convolve2d(xdB, np.expand_dims(benv, 1), "full")  # 2D convolution
    xenv = xenv[nhalf : nhalf + nsamp, :]
    yenv = convolve2d(ydB, np.expand_dims(benv, 1), "full")
    yenv = yenv[nhalf : nhalf + nsamp, :]

    # Subsample the LP filtered envelopes
    space = floor(fsamp / fsub)
    index = np.arange(0, nsamp, space)
    xLP = xenv[index, :]
    yLP = yenv[index, :]

    return xLP, yLP


def CepCoef(xdB, ydB, thrCep, thrNerve, nbasis):
    """
    Function to compute the cepstral correlation coefficients between the
    reference signal and the distorted signal log envelopes. The silence
    portions of the signals are removed prior to the calculation based on the
    envelope of the reference signal. For each time sample, the log spectrum
    in dB SL is fitted with a set of half-cosine basis functions. The
    cepstral coefficients then form the input to the cepstral correlation
    calculation.

    Args:
        xdB ():      subsampled reference signal envelope in dB SL in each band
        ydB ():	    subsampled distorted output signal envelope
        thrCep ():   threshold in dB SPL to include sample in calculation
        thrNerve (): additive noise RMS for IHC firing (in dB)
        nbasis    number of cepstral basis functions to use

    Returns:
        tuple: xcep cepstral coefficient matrix for the ref signal (nsamp,nbasis) and ycep  cepstral coefficient matrix
               for the output signal (nsamp,nbasis) each column is a separate basis function, from low to high

    James M. Kates, 23 April 2015.
    Gammawarp version to fit the basis functions, 11 February 2019.
    Additive noise for IHC firing rates, 24 April 2019.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Processing parameters
    nbands = xdB.shape[1]

    # Mel cepstrum basis functions
    freq = np.arange(0, nbasis)
    k = np.arange(0, nbands)
    cepm = np.zeros((nbands, nbasis))

    for nb in range(nbasis):
        basis = np.cos(freq[nb] * np.pi * k / (nbands - 1))
        cepm[:, nb] = basis / np.sqrt(np.sum(basis**2))

    # Find the reference segments that lie sufficiently above the quiescent rate
    xLinear = 10 ** (xdB / 20)  # Convert envelope dB to linear (specific loudness)
    xsum = np.sum(xLinear, 1) / nbands  # Proportional to loudness in sones
    xsum = 20 * np.log10(xsum)  # Convert back to dB (loudness in phons)
    index = np.where(xsum > thrCep)[0]  # Identify those segments above threshold
    nsamp = len(index)  # Number of segments above threshold

    # Exit if not enough segments above zero
    assert nsamp > 1, "Function ebm_CepCoef: Signal below threshold"

    # Remove the silent samples
    xdB = xdB[index, :]
    ydB = ydB[index, :]

    # Add low-level noise to provide IHC firing jitter
    xdB = AddNoise(xdB, thrNerve)
    ydB = AddNoise(ydB, thrNerve)

    # Compute the mel cepstrum coefficients using only those samples above threshold
    xcep = xdB @ cepm
    ycep = ydB @ cepm

    # Remove the average value from the cepstral coefficients. The cepstral cross-correlation
    # will thus be a cross-covariance, and there is no effect of the absolute signal level in dB.
    for n in range(nbasis):
        x = xcep[:, n]
        x = x - np.mean(x)
        xcep[:, n] = x
        y = ycep[:, n]
        y = y - np.mean(y)
        ycep[:, n] = y

    return xcep, ycep


def AddNoise(ydB, thrdB):
    """
    Function to add independent random Gaussian noise to the subsampled
    signal envelope in each auditory frequency band.

    Args:
    ydB      subsampled envelope in dB re:auditory threshold
    thrdB    additive noise RMS level (in dB)
    Level1   an input having RMS=1 corresponds to Level1 dB SPL

    Returns:
    zdB      envelope with threshold noise added, in dB re:auditory threshold

    James M. Kates, 23 April 2019.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Additive noise sequence
    # rng = np.random.default_rng()
    noise = thrdB * np.random.standard_normal(
        ydB.shape
    )  # Gaussian noise with RMS=1, then scaled

    # Add the noise to the signal envelope
    zdB = ydB + noise
    return zdB


def ModFilt(Xenv, Yenv, fsub):
    """
    Function to apply an FIR modulation filterbank to the reference envelope
    signals contained in matrix Xenv and the processed signal envelope
    signals in matrix Yenv. Each column in Xenv and Yenv is a separate filter
    band or cepstral coefficient basis function. The modulation filters use a
    lowpass filter for the lowest modulation rate, and complex demodulation
    followed by a lowpass filter for the remaining bands. The onset and
    offset transients are removed from the FIR convolutions to temporally
    align the modulation filter outputs.

    Args:
        Xenv (np.ndarray) : matrix containing the subsampled reference envelope values. Each
             column is a different frequency band or cepstral basis function
             arranged from low to high.
        Yenv (np.ndarray): matrix containing the subsampled processed envelope values
        fsub (): envelope sub-sampling rate in Hz

    Returns:
        tuple: Xmod a cell array containing the reference signal output of the
             modulation filterbank. Xmod is of size {nchan,nmodfilt} where
             nchan is the number of frequency channels or cepstral basis
             functions in Xenv, and nmodfilt is the number of modulation
             filters used in the analysis. Each cell contains a column vector
             of length nsamp, where nsamp is the number of samples in each
             envelope sequence contained in the columns of Xenv.
             Ymod cell array containing the processed signal output of the
             modulation filterbank.
             cf vector of the modulation rate filter center frequencies

    James M. Kates, 14 February 2019.
    Two matrix version of gwarp_ModFiltWindow, 19 February 2019.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """

    # Input signal properties
    nsamp = Xenv.shape[0]
    nchan = Xenv.shape[1]

    # Modulation filter band cf and edges, 10 bands
    # Band spacing uses the factor of 1.6 used by Dau
    cf = np.array([2, 6, 10, 16, 25, 40, 64, 100, 160, 256])  # Band center frequencies
    nmod = len(cf)
    edge = np.zeros(nmod + 1)
    edge[0:3] = [0, 4, 8]
    for k in range(3, nmod + 1):
        edge[k] = (cf[k - 1] ** 2) / edge[k - 1]

    # Allowable filters based on envelope subsampling rate
    fNyq = 0.5 * fsub
    index = edge < fNyq
    edge = edge[index]  # Filter upper bands edges less than Nyquist rate
    nmod = len(edge) - 1
    cf = cf[:nmod]

    # Assign FIR filter lengths. Setting t0=0.2 gives a filter Q of about 1.25
    # to match Ewert et al. (2002), and t0=0.33 corresponds to Q=2 (Dau et al 1997a).
    # Moritz et al. (2015) used t0=0.29. General relation Q=6.25*t0, compromise with
    # t0=0.24 which gives Q=1.5

    t0 = 0.24  # Filter length in seconds for the lowest modulation frequency band
    t = np.zeros(nmod)
    t[0] = t0
    t[1] = t0
    t[2:nmod] = t0 * cf[2] / cf[2:nmod]  # Constant-Q filters above 10 Hz
    nfir = 2 * np.floor(t * fsub / 2)  # Force even filter lengths in samples
    nhalf = nfir / 2

    # Design the family of lowpass windows
    b = []  # Filter coefficients, one filter impulse response per list element
    for k in range(nmod):
        bk = np.hanning(nfir[k] + 1)
        bk = bk / np.sum(bk)
        b.append(bk)

    # Pre-compute the cosine and sine arrays
    co = []  # cosine array, one frequency per list element
    si = []  # sine array, one frequency per list element
    n = np.arange(1, nsamp + 1)
    for k in range(nmod):
        if k == 0:
            co.append(1)
            si.append(0)
        else:
            co.append(np.sqrt(2) * np.cos(np.pi * n * cf[k] / fNyq))
            si.append(np.sqrt(2) * np.sin(np.pi * n * cf[k] / fNyq))

    # Convolve the input and output envelopes with the modulation filters
    Xmod = np.zeros((nchan, nmod, nsamp))
    Ymod = np.zeros((nchan, nmod, nsamp))
    for k in range(nmod):  # Loop over the modulation filters
        bk = b[k]  # Extract the lowpass filter impulse response
        nh = int(nhalf[k])  # Transient duration for the filter
        c = co[k]  # Cosine and sine for complex modulation
        s = si[k]
        for m in range(nchan):  # Loop over the input signal vectors
            # Reference signal
            x = Xenv[:, m]  # Extract the frequency or cepstral coefficient band
            u = convolve(
                (x * c - 1j * x * s), bk
            )  # Complex demodulation, then LP filter
            u = u[nh : nh + nsamp]  # Truncate the filter transients
            xfilt = (
                np.real(u) * c - np.imag(u) * s
            )  # Modulate back up to the carrier freq
            Xmod[m, k, :] = xfilt  # Save the filtered signal

            # Processed signal
            y = Yenv[:, m]  # Extract the frequency or cepstral coefficient band
            v = convolve(
                (y * c - 1j * y * s), bk
            )  # Complex demodulation, then LP filter
            v = v[nh : nh + nsamp]  # Truncate the filter transients
            yfilt = (
                np.real(v) * c - np.imag(v) * s
            )  # Modulate back up to the carrier freq
            Ymod[m, k, :] = yfilt  # Save the filtered signal

    return Xmod, Ymod, cf


def ModCorr(Xmod, Ymod):
    """
    Function to compute the cross-correlations between the input signal
    time-frequency envelope and the distortion time-frequency envelope. The
    cepstral coefficients or envelopes in each frequency band have been
    passed through the modulation filterbank using function ebm_ModFilt.

    Args:
        Xmod (np.array): cell array containing the reference signal output of the
             modulation filterbank. Xmod is of size {nchan,nmodfilt} where
             nchan is the number of frequency channels or cepstral basis
             functions in Xenv, and nmodfilt is the number of modulation
             filters used in the analysis. Each cell contains a column vector
             of length nsamp, where nsamp is the number of samples in each
             envelope sequence contained in the columns of Xenv.
        Ymod subsampled distorted output signal envelope

    Output:
        float: aveCM modulation correlations averaged over basis functions 2-6
             vector of size nmodfilt

    James M. Kates, 21 February 2019.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Processing parameters
    nchan = Xmod.shape[0]  # Number of basis functions
    nmod = Xmod.shape[1]  # Number of modulation filters
    small = 1e-30  # Zero threshold

    # Compute the cross-covariance matrix
    CM = np.zeros((nchan, nmod))
    for m in range(nmod):
        for j in range(nchan):
            # Index j gives the input reference band
            xj = Xmod[j, m]  # Input freq band j, modulation freq m
            xj = xj - np.mean(xj)
            xsum = np.sum(xj**2)
            # Processed signal band
            yj = Ymod[j, m]  # Processed freq band j, modulation freq m
            yj = yj - np.mean(yj)
            ysum = np.sum(yj**2)
            # Cross-correlate the reference and processed signals
            if (xsum < small) or (ysum < small):
                CM[j, m] = 0
            else:
                CM[j, m] = np.abs(np.sum(xj * yj)) / np.sqrt(xsum * ysum)

    aveCM = np.mean(CM[1:6], 0)

    return aveCM



"""Evaluate the enhanced signals using a combination of HASPI and HASQI metrics"""
import csv
import hashlib
import json
import logging
import pathlib

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.io import wavfile
from tqdm import tqdm

from clarity.enhancer.compressor import Compressor
from clarity.enhancer.nalr import NALR
from clarity.evaluator.haspi import haspi_v2_be
from clarity.evaluator.hasqi import hasqi_v2_be

logger = logging.getLogger(__name__)


def amplify_signal(signal, bin_audiogram, ear, enhancer, compressor):
    """Amplify signal for either left (l) or right (r) ear"""
    cfs = np.array(bin_audiogram["audiogram_cfs"])
    audiogram = np.array(bin_audiogram[f"audiogram_levels_{ear}"])
    nalr_fir, _ = enhancer.build(audiogram, cfs)
    out = enhancer.apply(nalr_fir, signal)
    out, _, _ = compressor.process(out)
    return out


def set_scene_seed(scene):
    """Set a seed that is unique for the given scene"""
    scene_encoded = hashlib.md5(scene.encode("utf-8")).hexdigest()
    scene_md5 = int(scene_encoded, 16) % (10**8)
    np.random.seed(scene_md5)


def compute_metric(metric, signal, ref, audiogram, fs_signal):
    """Compute HASPI or HASQI metric"""
    score = metric(
        xl=ref[:, 0],
        xr=ref[:, 1],
        yl=signal[:, 0],
        yr=signal[:, 1],
        fs_signal=fs_signal,
        audiogram_l=audiogram["audiogram_levels_l"],
        audiogram_r=audiogram["audiogram_levels_r"],
        audiogram_cfs=audiogram["audiogram_cfs"],
    )
    return score


class ResultsFile:
    """Class to write results to a CSV file"""

    def __init__(self, file_name):
        self.file_name = file_name

    def write_header(self):
        with open(self.file_name, "w", encoding="utf-8") as csv_f:
            csv_writer = csv.writer(
                csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
            )
            csv_writer.writerow(["scene", "listener", "combined", "haspi", "hasqi"])

    def add_result(self, scene, listener, score, haspi, hasqi):
        """Add a result to the CSV file"""

        logger.info(f"The combined score is {score} (haspi {haspi}, hasqi {hasqi})")

        with open(self.file_name, "a", encoding="utf-8") as csv_f:
            csv_writer = csv.writer(
                csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
            )
            csv_writer.writerow([scene, listener, str(score), str(haspi), str(hasqi)])


def make_scene_listener_list(scenes_listeners, small_test=False):
    """Make the list of scene-listener pairing to process"""
    scene_listener_pairs = [
        (scene, listener)
        for scene in scenes_listeners
        for listener in scenes_listeners[scene]
    ]

    # Can define a standard 'small_test' with just 1/15 of the data
    if small_test:
        scene_listener_pairs = scene_listener_pairs[::15]

    return scene_listener_pairs


@hydra.main(config_path=".", config_name="config")
def run_calculate_si(cfg: DictConfig) -> None:
    """Evaluate the enhanced signals using a combination of HASPI and HASQI metrics"""

    # Load listener data
    with open(cfg.path.scenes_listeners_file, "r", encoding="utf-8") as fp:
        scenes_listeners = json.load(fp)

    with open(cfg.path.listeners_file, "r", encoding="utf-8") as fp:
        listener_audiograms = json.load(fp)

    enhancer = NALR(**cfg.nalr)
    compressor = Compressor(**cfg.compressor)

    enhanced_folder = pathlib.Path("enhanced_signals")
    amplified_folder = pathlib.Path("amplified_signals")
    scenes_folder = pathlib.Path(cfg.path.scenes_folder)
    amplified_folder.mkdir(parents=True, exist_ok=True)

    # Make list of all scene listener pairs that will be run

    scene_listener_pairs = make_scene_listener_list(
        scenes_listeners, cfg.evaluate.small_test
    )

    results_file = ResultsFile("scores.csv")
    results_file.write_header()

    for scene, listener in tqdm(scene_listener_pairs):
        logger.info(f"Running evaluation: scene {scene}, listener {listener}")

        if cfg.evaluate.set_random_seed:
            set_scene_seed(scene)

        # Read signals

        fs_signal, signal = wavfile.read(
            enhanced_folder / f"{scene}_{listener}_enhanced.wav"
        )
        fs_ref_anechoic, ref_anechoic = wavfile.read(
            scenes_folder / f"{scene}_target_anechoic_CH1.wav"
        )
        fs_ref_target, ref_target = wavfile.read(
            scenes_folder / f"{scene}_target_CH1.wav"
        )
        ref_anechoic = ref_anechoic / 32768.0
        ref_target = ref_target / 32768.0

        assert fs_ref_anechoic == fs_ref_target == fs_signal == cfg.nalr.fs

        # amplify left and right ear signals
        audiogram = listener_audiograms[listener]

        out_l = amplify_signal(signal[:, 0], audiogram, "l", enhancer, compressor)
        out_r = amplify_signal(signal[:, 1], audiogram, "r", enhancer, compressor)
        amplified = np.stack([out_l, out_r], axis=1)

        if cfg.soft_clip:
            amplified = np.tanh(amplified)

        wavfile.write(
            amplified_folder / f"{scene}_{listener}_HA-output.wav",
            fs_signal,
            amplified.astype(np.float32),
        )

        # Evaluate the amplified signal

        rms_target = np.mean(ref_target**2, axis=0) ** 0.5
        rms_anechoic = np.mean(ref_anechoic**2, axis=0) ** 0.5
        ref = ref_anechoic * rms_target / rms_anechoic

        haspi_score = compute_metric(haspi_v2_be, amplified, ref, audiogram, fs_signal)
        hasqi_score = compute_metric(hasqi_v2_be, amplified, ref, audiogram, fs_signal)
        score = 0.5 * (hasqi_score + haspi_score)

        results_file.add_result(
            scene, listener, score=score, haspi=haspi_score, hasqi=hasqi_score
        )


if __name__ == "__main__":
    run_calculate_si()  # noqa



"""Implementation of the MSBG hearing loss model."""
import logging
import math

import numpy as np
from scipy import interpolate, signal
from scipy.signal import firwin, lfilter

from clarity.evaluator.msbg.cochlea import Cochlea
from clarity.evaluator.msbg.msbg_utils import (
    DF_ED,
    FF_ED,
    HZ,
    ITU_ERP_DRP,
    ITU_HZ,
    MIDEAR,
    firwin2,
    gen_eh2008_speech_noise,
    gen_tone,
    measure_rms,
)

# Cut off frequency of low-pass filter at end of simulations:
# prevents possible excessive processing noise at high frequencies.
UPPER_CUTOFF_HZ = 18000


class Ear:
    """Representation of a pairs of ears."""

    def __init__(
        self, src_pos="ff", sample_frequency: float = 44100.0, equiv_0db_spl=100, ahr=20
    ) -> None:
        """
        Constructor for the Ear class.
        Args:
            src_pos (str): Position of the source.
            sample_frequency (float): sample frequency.
            equiv_0db_spl (): ???
            ahr (): ???
        """
        self.sample_frequency = sample_frequency
        self.src_correction = self.get_src_correction(src_pos)
        self.equiv_0db_spl = equiv_0db_spl
        self.ahr = ahr
        self.cochlea = None

    def set_audiogram(self, audiogram):
        """Set the audiogram to be used."""
        if np.max(audiogram.levels[audiogram.levels is not None]) > 80:
            logging.warning(
                "Impairment too severe: Suggest you limit audiogram max to 80-90 dB HL, \
                otherwise things go wrong/unrealistic."
            )
        self.cochlea = Cochlea(audiogram=audiogram)

    @staticmethod
    def get_src_correction(src_pos: str) -> np.array:
        """Select relevant external field to eardrum correction.

        Args:
            src_pos (str): Position of src. One of ff, df or ITU

        """
        if src_pos == "ff":
            src_correction = FF_ED
        elif src_pos == "df":
            src_correction = DF_ED
        elif src_pos == "ITU":  # transfer to same grid
            field = interpolate.interp1d(ITU_HZ, ITU_ERP_DRP, kind="linear")
            src_correction = field(HZ)
        else:
            logging.error(
                f"Invalid src position ({src_pos}). Must be one of ff, df or ITU"
            )
            raise ValueError("Invalid src position")
        return src_correction

    @staticmethod
    def src_to_cochlea_filt(
        ip_sig: np.ndarray, src_correction: str, sample_frequency: int, backward=False
    ) -> np.ndarray:
        """Simulate middle and outer ear transfer functions.

        Made more general, Mar2012, to include diffuse field as well as ITU reference points,
        that were included in DOS-versions of recruitment simulator, released ca 1999-2001,
        and on hearing group website, Mar2012 variable [src_pos] takes one of 3
        values: 'ff', 'df' and 'ITU' free-field to cochlea filter forwards or backward direction,
        depends on 'backward' switch. NO LONGER via 2 steps. ff to eardrum and then via middle ear:
        use same length FIR 5-12-97.

        Args:
            ip_sig (ndarray): signal to process
            src_correction (str): correction to make for src position
            sample_frequency (int): sampling frequency
            backward (bool, optional): if true then cochlea to src (default: False)

        Returns:
            np.ndarray: the processed signal

        """
        logging.info("performing outer/middle ear corrections")

        # make sure that response goes only up to sample_frequency/2
        nyquist = int(sample_frequency / 2)
        ixf_useful = np.nonzero(HZ < nyquist)

        hz_used = HZ[ixf_useful]
        hz_used = np.append(hz_used, nyquist)

        # sig from free field to cochlea: 0 dB at 1kHz
        correction = src_correction - MIDEAR
        field = interpolate.interp1d(HZ, correction, kind="linear")
        last_correction = field(nyquist)  # generate synthetic response at Nyquist

        correction_used = np.append(correction[ixf_useful], last_correction)
        if backward:  # ie. coch->src rather than src->coch
            correction_used = -correction_used
        correction_used = np.power(10, (0.05 * correction_used))

        correction_used = correction_used.flatten()
        # Create filter with 23 msec window to do reasonable job down to about 100 Hz
        # Scales with fs, falls over with longer windows in fir2 in original MATLAB version
        n_wdw = 2 * math.floor((sample_frequency / 16e3) * 368 / 2)
        hz_used = hz_used / nyquist

        b = firwin2(n_wdw + 1, hz_used.flatten(), correction_used, window=("kaiser", 4))
        op_sig = signal.lfilter(b, 1, ip_sig)

        return op_sig

    def make_calibration_signal(self, ref_rms_db):
        """Add the calibration signal to the start of the signal.

        Args:
            ref_rms_db (ndarray): input signal

        Returns:
            ndarray: the processed signal

        """
        # Calibration noise and tone with same RMS as original speech,
        # Tone at nearest channel centre frequency to 500 Hz
        # For testing, ref_rms_dB must be equal to -31.2

        noise_burst = gen_eh2008_speech_noise(
            duration=2, sample_frequency=self.sample_frequency, level=ref_rms_db
        )
        tone_burst = gen_tone(
            freq=520,
            duration=0.5,
            sample_frequency=self.sample_frequency,
            level=ref_rms_db,
        )
        silence = np.zeros(int(0.05 * self.sample_frequency))  # 50 ms duration
        return (
            np.concatenate((silence, tone_burst, silence, noise_burst, silence)),
            silence,
        )

    @staticmethod
    def array_to_list(chans: np.ndarray) -> np.ndarray:
        """Convert signal into a list of 1-D arrays.

        Args:
            signal (np.ndarray) Signal to be converted, can be 1-D already.

        Returns:
            np.ndarray: A list of 1-D arrays.
        """
        if len(chans.shape) == 1:
            chans = chans[..., np.newaxis]
        return [chans[:, i] for i in range(chans.shape[1])]

    def process(self, signal: np.ndarray, add_calibration: bool = False) -> np.ndarray:
        """Run the hearing loss simulation.

        Args:
            signal (ndarray): signal to process, shape either N, Nx1, Nx2
            add_calibration (bool): prepend calibration tone and speech-shaped noise
                (default: False)

        Returns:
            np.ndarray: the processed signal

        """

        sample_frequency = 44100  # This is the only sampling frequency that can be used
        if sample_frequency != self.sample_frequency:
            logging.error(
                "Warning: only a sampling frequency of 44.1kHz can be used by MSBG."
            )

        logging.info("Processing {len(chans)} samples")

        # Get single channel array and convert to list
        signal = Ear.array_to_list(signal)

        # Need to know file RMS, and then call that a certain level in SPL:
        # needs some form of pre-measuring.
        level_resample_frequency = 10 * np.log10(np.mean(np.array(signal) ** 2))

        equiv_0db_spl = self.equiv_0db_spl + self.ahr

        level_db_spl = equiv_0db_spl + level_resample_frequency
        calib_db_spl = level_db_spl
        target_spl = level_db_spl
        ref_rms_db = calib_db_spl - equiv_0db_spl

        # Measure RMS where 3rd arg is dB_rel_rms (how far below)
        calculated_rms, idx, _rel_db_thresh, _active = measure_rms(
            signal[0], sample_frequency, -12
        )

        # Rescale input data and check level after rescaling
        # This is to ensure that the following processing steps are applied correctly
        change_db = target_spl - (equiv_0db_spl + 20 * np.log10(calculated_rms))
        signal = [x * np.power(10, 0.05 * change_db) for x in signal]
        new_rms_db = equiv_0db_spl + 10 * np.log10(
            np.mean(np.power(signal[0][idx], 2.0))
        )
        logging.info(
            "Rescaling: "
            f"leveldBSPL was {level_db_spl:3.1f} dB SPL, now {new_rms_db:3.1f} dB SPL. "
            f" Target SPL is {target_spl:3.1f} dB SPL."
        )

        # Add calibration signal at target SPL dB
        if add_calibration is True:
            # if self.calibration_signal is None:
            #     self.calibration_signal = self.make_calibration_signal(ref_rms_db)
            # signal = [
            #     np.concatenate(
            #         (self.calibration_signal[0], x, self.calibration_signal[1])
            #     )
            #     for x in signal
            # ]
            calibration_signal = self.make_calibration_signal(ref_rms_db)
            signal = [
                np.concatenate((calibration_signal[0], x, calibration_signal[1]))
                for x in signal
            ]

        # Transform from src pos to cochlea, simulate cochlea, transform back to src pos
        signal = [
            Ear.src_to_cochlea_filt(x, self.src_correction, sample_frequency)
            for x in signal
        ]
        if self.cochlea is not None:
            signal = [self.cochlea.simulate(x, equiv_0db_spl) for x in signal]
        signal = [
            Ear.src_to_cochlea_filt(
                x, self.src_correction, sample_frequency, backward=True
            )
            for x in signal
        ]

        # Implement low-pass filter at top end of audio range: flat to Cutoff freq, tails
        # below -80 dB. Suitable lpf for signals later converted to MP3, flat to 15 kHz.
        # Small window to design low-pass FIR, to cut off high freq processing noise
        # low-pass to something sensible, prevents exaggeration of > 15 kHz
        winlen = 2 * math.floor(0.0015 * sample_frequency) + 1
        lpf44d1 = firwin(
            winlen, UPPER_CUTOFF_HZ / int(sample_frequency / 2), window=("kaiser", 8)
        )
        signal = [lfilter(lpf44d1, 1, x) for x in signal]

        return signal



"""
Adopted from https://github.com/kaituoxu/Conv-TasNet
"""
import math
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

EPS = 1e-8


def overlap_and_add(signal, frame_step, device):
    """Reconstructs a signal from a framed representation.

    Adds potentially overlapping frames of a signal with shape
    `[..., frames, frame_length]`, offsetting subsequent frames by `frame_step`.
    The resulting tensor has shape `[..., output_size]` where

        output_size = (frames - 1) * frame_step + frame_length

    Args:
        signal: A [..., frames, frame_length] Tensor. All dimensions may be unknown,
            and rank must be at least 2.
        frame_step: An integer denoting overlap offsets. Must be less than or equal to frame_length.
        device: Whether to use 'cpu' or 'cuda' for processing.

    Returns:
        A Tensor with shape [..., output_size] containing the overlap-added frames of
            signal's inner-most two dimensions.
        output_size = (frames - 1) * frame_step + frame_length

    Based on https://github.com/tensorflow/tensorflow/blob/r1.12/tensorflow/contrib/signal/python/ops/reconstruction_ops.py
    """
    outer_dimensions = signal.size()[:-2]
    frames, frame_length = signal.size()[-2:]

    subframe_length = math.gcd(frame_length, frame_step)  # gcd=Greatest Common Divisor
    subframe_step = frame_step // subframe_length
    subframes_per_frame = frame_length // subframe_length
    output_size = frame_step * (frames - 1) + frame_length
    output_subframes = output_size // subframe_length

    subframe_signal = signal.view(*outer_dimensions, -1, subframe_length)

    frame = torch.arange(0, output_subframes).unfold(
        0, subframes_per_frame, subframe_step
    )
    frame = signal.new_tensor(frame, device=device).long()  # signal may in GPU or CPU
    frame = frame.contiguous().view(-1)

    result = signal.new_zeros(
        *outer_dimensions, output_subframes, subframe_length, device=device
    )
    result.index_add_(-2, frame, subframe_signal)
    result = result.view(*outer_dimensions, -1)
    return result


class ConvTasNet(nn.Module):
    def __init__(
        self,
        N_spec,
        N_spat,
        L,
        B,
        H,
        P,
        X,
        R,
        C,
        num_channels,
        norm_type="cLN",
        causal=False,
        mask_nonlinear="relu",
        device=None,
    ):
        """
        Args:
            N_spec: Number of filters in spectral autoencoder
            N_spat: Number of filters in spatial autoencoder
            L: Length of the filters (in samples)
            B: Number of channels in bottleneck 1 × 1-conv block
            H: Number of channels in convolutional blocks
            P: Kernel size in convolutional blocks
            X: Number of convolutional blocks in each repeat
            R: Number of repeats
            C: Number of speakers
            norm_type: BN, gLN, cLN
            causal: causal or non-causal
            mask_nonlinear: use which non-linear function to generate mask
        """
        super().__init__()

        # Hyper-parameter
        (
            self.N_spec,
            self.N_spat,
            self.L,
            self.B,
            self.H,
            self.P,
            self.X,
            self.R,
            self.C,
        ) = (N_spec, N_spat, L, B, H, P, X, R, C)
        self.num_channels = num_channels
        self.norm_type = norm_type
        self.causal = causal
        self.mask_nonlinear = mask_nonlinear
        # Components
        self.spectral_encoder = SpectralEncoder(L, N_spec)
        self.spatial_encoder = SpatialEncoder(L, N_spat, num_channels)

        self.separator = TemporalConvNet(
            N_spec, N_spat, B, H, P, X, R, C, norm_type, causal, mask_nonlinear
        )
        self.decoder = Decoder(N_spec, L, device)
        # init
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_normal_(p)

    def forward(self, mixture):
        """
        Args:
            mixture: [M, T], M is batch size, T is #samples

        Returns:
            est_source: [M, C, T]
        """
        mixture_spectral = self.spectral_encoder(
            torch.narrow(mixture, 1, 0, 1).squeeze(1)
        )
        mixture_spatial = self.spatial_encoder(mixture)

        mixture_w = torch.cat((mixture_spectral, mixture_spatial), 1)
        est_mask = self.separator(mixture_w)
        est_source = self.decoder(mixture_spectral, est_mask)

        # T changed after conv1d in encoder, fix it here
        T_origin = mixture.size(-1)
        T_conv = est_source.size(-1)
        est_source = F.pad(est_source, (0, T_origin - T_conv))
        return est_source


class SpectralEncoder(nn.Module):
    """Estimation of the nonnegative mixture weight by a 1-D conv layer."""

    def __init__(self, L, N):
        super().__init__()
        # Hyper-parameter
        self.L, self.N = L, N
        # Components
        # 50% overlap
        self.conv1d_U = nn.Conv1d(1, N, kernel_size=L, stride=L // 2, bias=False)

    def forward(self, mixture):
        """
        Args:
            mixture: [M, T], M is batch size, T is #samples

        Returns:
            mixture_w: [M, N, K], where K = (T-L)/(L/2)+1 = 2T/L-1
        """
        mixture = torch.unsqueeze(mixture, 1)  # [M, 1, T]
        mixture_w = F.relu(self.conv1d_U(mixture))  # [M, N, K]
        return mixture_w


class SpatialEncoder(nn.Module):
    """Estimation of the nonnegative mixture weight by a 1-D conv layer."""

    def __init__(self, L, N, num_channels):
        super(SpatialEncoder, self).__init__()
        # Hyper-parameter
        self.L, self.N = L, N
        # Components
        # 50% overlap
        self.conv1d_U = nn.Conv1d(
            num_channels, N, kernel_size=L, stride=L // 2, bias=False
        )

    def forward(self, mixture):
        """
        Args:
            mixture: [M, num_channels, T], M is batch size, T is #samples
        Returns:
            mixture_w: [M, N, K], where K = (T-L)/(L/2)+1 = 2T/L-1
        """
        mixture_w = F.relu(self.conv1d_U(mixture))  # [M, N, K]
        return mixture_w


class Decoder(nn.Module):
    def __init__(self, N, L, device: Optional[str] = None):
        super().__init__()
        # device for overlap_and add
        if device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self.device = device
        # Hyper-parameter
        self.N, self.L = N, L
        # Components
        self.basis_signals = nn.Linear(N, L, bias=False)

    def forward(self, mixture_w, est_mask):
        """
        Args:
            mixture_w: [M, N, K]
            est_mask: [M, C, N, K]

        Returns:
            est_source: [M, C, T]
        """
        # D = W * M
        source_w = torch.unsqueeze(mixture_w, 1) * est_mask  # [M, C, N, K]
        source_w = torch.transpose(source_w, 2, 3)  # [M, C, K, N]
        # S = DV
        est_source = self.basis_signals(source_w)  # [M, C, K, L]
        est_source = overlap_and_add(
            est_source, self.L // 2, device=self.device
        )  # M x C x T
        return est_source


class TemporalConvNet(nn.Module):
    def __init__(
        self,
        N_spec,
        N_spat,
        B,
        H,
        P,
        X,
        R,
        C,
        norm_type="gLN",
        causal=False,
        mask_nonlinear="relu",
    ):
        """
        Args:
            N: Number of filters in autoencoder
            B: Number of channels in bottleneck 1 × 1-conv block
            H: Number of channels in convolutional blocks
            P: Kernel size in convolutional blocks
            X: Number of convolutional blocks in each repeat
            R: Number of repeats
            C: Number of speakers
            norm_type: BN, gLN, cLN
            causal: causal or non-causal
            mask_nonlinear: use which non-linear function to generate mask
        """
        super().__init__()
        # Hyper-parameter
        self.C, self.N_spec = C, N_spec
        self.mask_nonlinear = mask_nonlinear
        # Components
        # [M, N, K] -> [M, N, K]
        layer_norm = ChannelwiseLayerNorm(N_spec + N_spat)
        # [M, N, K] -> [M, B, K]
        bottleneck_conv1x1 = nn.Conv1d(N_spec + N_spat, B, 1, bias=False)
        # [M, B, K] -> [M, B, K]
        repeats = []
        for _r in range(R):
            blocks = []
            for x in range(X):
                dilation = 2**x
                padding = (P - 1) * dilation if causal else (P - 1) * dilation // 2
                blocks += [
                    TemporalBlock(
                        B,
                        H,
                        P,
                        stride=1,
                        padding=padding,
                        dilation=dilation,
                        norm_type=norm_type,
                        causal=causal,
                    )
                ]
            repeats += [nn.Sequential(*blocks)]
        temporal_conv_net = nn.Sequential(*repeats)
        # [M, B, K] -> [M, C*N, K]
        mask_conv1x1 = nn.Conv1d(B, C * N_spec, 1, bias=False)
        # Put together
        self.network = nn.Sequential(
            layer_norm, bottleneck_conv1x1, temporal_conv_net, mask_conv1x1
        )

    def forward(self, mixture_w):
        """
        Keep this API same with TasNet

        Args:
            mixture_w: [M, N, K], M is batch size

        Returns:
            est_mask: [M, C, N, K]
        """
        M, _N, K = mixture_w.size()
        score = self.network(mixture_w)  # [M, N, K] -> [M, C*N, K]
        score = score.view(M, self.C, self.N_spec, K)  # [M, C*N, K] -> [M, C, N, K]
        if self.mask_nonlinear == "softmax":
            est_mask = F.softmax(score, dim=1)
        elif self.mask_nonlinear == "relu":
            est_mask = F.relu(score)
        elif self.mask_nonlinear == "tanh":
            est_mask = F.tanh(score)
        else:
            raise ValueError("Unsupported mask non-linear function")
        return est_mask


class TemporalBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
        dilation,
        norm_type="gLN",
        causal=False,
    ):
        super().__init__()
        # [M, B, K] -> [M, H, K]
        conv1x1 = nn.Conv1d(in_channels, out_channels, 1, bias=False)
        prelu = nn.PReLU()
        norm = chose_norm(norm_type, out_channels)
        # [M, H, K] -> [M, B, K]
        dsconv = DepthwiseSeparableConv(  # pylint: disable=W1114
            out_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            norm_type,
            causal,
        )
        # Put together
        self.net = nn.Sequential(conv1x1, prelu, norm, dsconv)

    def forward(self, x):
        """
        Args:
            x: [M, B, K]
        Returns:
            [M, B, K]
        """
        residual = x
        out = self.net(x)
        # TODO: when P = 3 here works fine, but when P = 2 maybe need to pad?
        return out + residual  # look like w/o F.relu is better than w/ F.relu
        # return F.relu(out + residual)


class DepthwiseSeparableConv(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
        dilation,
        norm_type="gLN",
        causal=False,
    ):
        super().__init__()
        # Use `groups` option to implement depthwise convolution
        # [M, H, K] -> [M, H, K]
        depthwise_conv = nn.Conv1d(
            in_channels,
            in_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=in_channels,
            bias=False,
        )
        if causal:
            chomp = Chomp1d(padding)
        prelu = nn.PReLU()
        norm = chose_norm(norm_type, in_channels)
        # [M, H, K] -> [M, B, K]
        pointwise_conv = nn.Conv1d(in_channels, out_channels, 1, bias=False)
        # Put together
        if causal:
            self.net = nn.Sequential(depthwise_conv, chomp, prelu, norm, pointwise_conv)
        else:
            self.net = nn.Sequential(depthwise_conv, prelu, norm, pointwise_conv)

    def forward(self, x):
        """
        Args:
            x: [M, H, K]
        Returns:
            result: [M, B, K]
        """
        return self.net(x)


class Chomp1d(nn.Module):
    """To ensure the output length is the same as the input."""

    def __init__(self, chomp_size):
        super().__init__()
        self.chomp_size = chomp_size

    def forward(self, x):
        """
        Args:
            x: [M, H, Kpad]
        Returns:
            [M, H, K]
        """
        return x[:, :, : -self.chomp_size].contiguous()


def chose_norm(norm_type, channel_size):
    """The input of normlization will be (M, C, K), where M is batch size,
    C is channel size and K is sequence length.

    Args:
        normn_type ():
        channel_size ():
    Returns:

    """
    if norm_type == "gLN":
        return GlobalLayerNorm(channel_size)
    if norm_type == "cLN":
        return ChannelwiseLayerNorm(channel_size)
    # norm_type == "BN":
    # # Given input (M, C, K), nn.BatchNorm1d(C) will accumulate statics
    # along M and K, so this BN usage is right.
    return nn.BatchNorm1d(channel_size)


# TODO: Use nn.LayerNorm to impl cLN to speed up
class ChannelwiseLayerNorm(nn.Module):
    """Channel-wise Layer Normalization (cLN)"""

    def __init__(self, channel_size):
        super().__init__()
        self.gamma = nn.Parameter(torch.Tensor(1, channel_size, 1))  # [1, N, 1]
        self.beta = nn.Parameter(torch.Tensor(1, channel_size, 1))  # [1, N, 1]
        self.reset_parameters()

    def reset_parameters(self):
        self.gamma.data.fill_(1)
        self.beta.data.zero_()

    def forward(self, y):
        """
        Args:
            y: [M, N, K], M is batch size, N is channel size, K is length

        Returns:
            cLN_y: [M, N, K]
        """
        mean = torch.mean(y, dim=1, keepdim=True)  # [M, 1, K]
        var = torch.var(y, dim=1, keepdim=True, unbiased=False)  # [M, 1, K]
        cLN_y = self.gamma * (y - mean) / torch.pow(var + EPS, 0.5) + self.beta
        return cLN_y


class GlobalLayerNorm(nn.Module):
    """Global Layer Normalization (gLN)"""

    def __init__(self, channel_size):
        super().__init__()
        self.gamma = nn.Parameter(torch.Tensor(1, channel_size, 1))  # [1, N, 1]
        self.beta = nn.Parameter(torch.Tensor(1, channel_size, 1))  # [1, N, 1]
        self.reset_parameters()

    def reset_parameters(self):
        self.gamma.data.fill_(1)
        self.beta.data.zero_()

    def forward(self, y):
        """
        Args:
            y: [M, N, K], M is batch size, N is channel size, K is length

        Returns:
            gLN_y: [M, N, K]
        """
        # TODO: in torch 1.0, torch.mean() support dim list
        mean = y.mean(dim=1, keepdim=True).mean(dim=2, keepdim=True)  # [M, 1, 1]
        var = (
            (torch.pow(y - mean, 2)).mean(dim=1, keepdim=True).mean(dim=2, keepdim=True)
        )
        gLN_y = self.gamma * (y - mean) / torch.pow(var + EPS, 0.5) + self.beta
        return gLN_y



import logging
import math
import os

import numpy as np
import soundfile
from scipy.signal import convolve
from soundfile import SoundFile

from clarity.data.utils import better_ear_speechweighted_snr, pad, sum_signals


class Renderer:
    """
    SceneGenerator of CEC1 training and development sets. The render() function generates all
    simulated signals for each scene given the parameters specified in the
    metadata/scenes.train.json or metadata/scenes.dev.json file.
    """

    def __init__(
        self,
        input_path,
        output_path,
        num_channels=1,
        fs=44100,
        ramp_duration=0.5,
        tail_duration=0.2,
        pre_duration=2.0,
        post_duration=1.0,
        test_nbits=16,
    ):

        self.input_path = input_path
        self.output_path = output_path
        self.fs = fs
        self.ramp_duration = ramp_duration
        self.n_tail = int(tail_duration * fs)
        self.pre_duration = pre_duration
        self.post_duration = post_duration
        self.test_nbits = test_nbits

        if num_channels == 0:
            # This will only generate the initial target, masker and anechoic target signal
            self.channels = []
        else:
            # ... as above plus N hearing aid input channels plus 'channel 0' (the eardrum signal).
            # e.g. num_channel = 2  => channels [1, 2, 0]
            self.channels = list(range(1, num_channels + 1)) + [0]

    def read_signal(
        self, filename, offset=0, nsamples=-1, nchannels=0, offset_is_samples=False
    ):
        """Read a wavefile and return as numpy array of floats.

        Args:
            filename (string): Name of file to read
            offset (int, optional): Offset in samples or seconds (from start). Defaults to 0.
            nchannels: expected number of channel (default: 0 = any number OK)
            offset_is_samples (bool): measurement units for offset (default: False)

        Returns:
            ndarray: audio signal
        """
        try:
            wave_file = SoundFile(filename)
        except Exception as e:
            # Ensure incorrect error (24 bit) is not generated
            raise Exception(f"Unable to read {filename}.") from e

        if nchannels not in (0, wave_file.channels):
            raise Exception(
                f"Wav file ({filename}) was expected to have {nchannels} channels."
            )

        if wave_file.samplerate != self.fs:
            raise Exception(f"Sampling rate is not {self.fs} for filename {filename}.")

        if not offset_is_samples:  # Default behaviour
            offset = int(offset * wave_file.samplerate)

        if offset != 0:
            wave_file.seek(offset)

        x = wave_file.read(frames=nsamples)
        return x

    def write_signal(self, filename, x, fs, floating_point=True):
        """Write a signal as fixed or floating point wav file."""

        if fs != self.fs:
            logging.warning("Sampling rate mismatch: %s with sr=%s.", filename, fs)
            # raise ValueError("Sampling rate mismatch")

        if floating_point is False:
            if self.test_nbits == 16:
                subtype = "PCM_16"
                # If signal is float and we want int16
                x *= 32768
                x = x.astype(np.dtype("int16"))
                assert np.max(x) <= 32767 and np.min(x) >= -32768
            elif self.test_nbits == 24:
                subtype = "PCM_24"
        else:
            subtype = "FLOAT"

        soundfile.write(filename, x, fs, subtype=subtype)

    def apply_ramp(self, x, dur):
        """Apply half cosine ramp into and out of signal

        dur - ramp duration in seconds
        """
        ramp = np.cos(np.linspace(math.pi, 2 * math.pi, int(self.fs * dur)))
        ramp = (ramp + 1) / 2
        y = np.array(x)
        y[0 : len(ramp)] *= ramp
        y[-len(ramp) :] *= ramp[::-1]
        return y

    def apply_brir(self, signal, brir):
        """Convolve a signal with a BRIR.

        Args:
            signal (ndarray): The mono or stereo signal stored as array of floats
            brir (ndarray): The binaural room impulse response stored a 2xN array of floats
            n_tail (int): Truncate output to input signal length + n_tail

        Returns:
            ndarray: The convolved signals

        """
        output_len = len(signal) + self.n_tail
        brir = np.squeeze(brir)

        if len(np.shape(signal)) == 1 and len(np.shape(brir)) == 2:
            signal_l = convolve(signal, brir[:, 0], mode="full", method="fft")
            signal_r = convolve(signal, brir[:, 1], mode="full", method="fft")
        elif len(np.shape(signal)) == 2 and len(np.shape(brir)) == 2:
            signal_l = convolve(signal[:, 0], brir[:, 0], mode="full", method="fft")
            signal_r = convolve(signal[:, 1], brir[:, 1], mode="full", method="fft")
        else:
            logging.error("Signal does not have the required shape.")
        output = np.vstack([signal_l, signal_r]).T
        return output[0:output_len, :]

    def compute_snr(self, target, noise, pre_samples=0, post_samples=-1):
        """Return the SNR.
        Take the overlapping segment of the noise and get the speech-weighted
        better ear SNR. (Note, SNR is a ratio -- not in dB.)
        """

        pre_samples = int(self.fs * self.pre_duration)
        post_samples = int(self.fs * self.post_duration)

        segment_target = target[pre_samples:-post_samples]
        segment_noise = noise[pre_samples:-post_samples]
        assert len(segment_target) == len(segment_noise)

        snr = better_ear_speechweighted_snr(segment_target, segment_noise)

        return snr

    def render(
        self,
        target,
        noise_type,
        interferer,
        room,
        scene,
        offset,
        snr_dB,
        dataset,
        pre_samples=88200,
        post_samples=44100,
    ):

        brir_stem = f"{self.input_path}/{dataset}/rooms/brir/brir_{room}"
        anechoic_brir_stem = f"{self.input_path}/{dataset}/rooms/brir/anech_brir_{room}"
        target_fn = f"{self.input_path}/{dataset}/targets/{target}.wav"
        interferer_fn = (
            f"{self.input_path}/{dataset}/interferers/{noise_type}/{interferer}.wav"
        )

        target = self.read_signal(target_fn)
        target = np.pad(target, [(pre_samples, post_samples)])

        interferer = self.read_signal(
            interferer_fn, offset=offset, nsamples=len(target), offset_is_samples=True
        )

        if len(target) != len(interferer):
            logging.debug("Target and interferer have different lengths")

        # Apply 500ms half-cosine ramp
        interferer = self.apply_ramp(interferer, dur=self.ramp_duration)

        prefix = f"{self.output_path}/{scene}"
        outputs = [
            (f"{prefix}_target.wav", target),
            (f"{prefix}_interferer.wav", interferer),
        ]

        snr_ref = None
        for channel in self.channels:
            # Load scene BRIRs
            target_brir_fn = f"{brir_stem}_t_CH{channel}.wav"
            interferer_brir_fn = f"{brir_stem}_i1_CH{channel}.wav"
            target_brir = self.read_signal(target_brir_fn)
            interferer_brir = self.read_signal(interferer_brir_fn)

            # Apply the BRIRs
            target_at_ear = self.apply_brir(target, target_brir)
            interferer_at_ear = self.apply_brir(interferer, interferer_brir)

            # Scale interferer to obtain SNR specified in scene description
            logging.info("Scaling interferer to obtain mixture SNR = %s dB.", snr_dB)

            if snr_ref is None:
                # snr_ref computed for first channel in the list and then
                # same scaling applied to all
                snr_ref = self.compute_snr(
                    target_at_ear,
                    interferer_at_ear,
                    pre_samples=pre_samples,
                    post_samples=post_samples,
                )
                logging.debug("Using channel %s as reference.", channel)

            # Apply snr_ref reference scaling to get 0 dB and then scale to target snr_dB
            interferer_at_ear = interferer_at_ear * snr_ref
            interferer_at_ear = interferer_at_ear * 10 ** ((-snr_dB) / 20)

            # Sum target and scaled and ramped interferer
            signal_at_ear = sum_signals([target_at_ear, interferer_at_ear])
            outputs.extend(
                [
                    (f"{prefix}_mixed_CH{channel}.wav", signal_at_ear),
                    (f"{prefix}_target_CH{channel}.wav", target_at_ear),
                    (f"{prefix}_interferer_CH{channel}.wav", interferer_at_ear),
                ]
            )

        if self.channels == []:
            target_brir_fn = f"{brir_stem}_t_CH0.wav"
            target_brir = self.read_signal(target_brir_fn)

        # Construct the anechoic target reference signal
        anechoic_brir_fn = (
            f"{anechoic_brir_stem}_t_CH1.wav"  # CH1 used for the anechoic signal
        )
        anechoic_brir = self.read_signal(anechoic_brir_fn)
        # Padding the anechoic brir very inefficient but keeps it simple
        anechoic_brir_pad = pad(anechoic_brir, len(target_brir))
        target_anechoic = self.apply_brir(target, anechoic_brir_pad)

        outputs.append((f"{prefix}_target_anechoic.wav", target_anechoic))

        # Write all output files
        for (filename, signal) in outputs:
            self.write_signal(filename, signal, self.fs)


def check_scene_exists(scene, output_path, num_channels):
    """Checks correct dataset directory for full set of pre-existing files.

    Args:
        scene (dict): dictionary defining the scene to be generated.

    Returns:
        status: boolean value indicating whether scene signals exist
            or do not exist.

    """
    channels = []
    if num_channels == 0:
        # This will only generate the initial target, masker and anechoic target signal
        pass
    else:
        # ... as above plus N hearing aid input channels plus 'channel 0' (the eardrum signal).
        # e.g. num_channel = 2  => channels [1, 2, 0]
        channels = list(range(1, num_channels + 1)) + [0]

    pattern = f"{output_path}/{scene['scene']}"
    files_to_check = [
        f"{pattern}_target.wav",
        f"{pattern}_target_anechoic.wav",
        f"{pattern}_interferer.wav",
    ]
    for ch in channels:
        files_to_check.extend(
            [
                f"{pattern}_mixed_CH{ch}.wav",
                f"{pattern}_interferer_CH{ch}.wav",
                f"{pattern}_target_CH{ch}.wav",
            ]
        )

    scene_exists = True
    for filename in files_to_check:
        scene_exists = scene_exists and os.path.exists(filename)
    return scene_exists



from clarity.evaluator.haspi import eb


def hasqi_v2(x, fx, y, fy, HL, eq=1, level1=65):
    """
    Function to compute the HASQI version 2 quality index using the
    auditory model followed by computing the envelope cepstral
    correlation and BM vibration average short-time coherence signals.
    The reference signal presentation level for NH listeners is assumed
    to be 65 dB SPL. The same model is used for both normal and
    impaired hearing.

    Arguments:
    x			Clear input reference speech signal with no noise or distortion.
              If a hearing loss is specified, NAL-R equalization is optional
    fx        Sampling rate in Hz for signal x
    y			Output signal with noise, distortion, HA gain, and/or processing.
    fy        Sampling rate in Hz for signal y.
    HL		(1,6) vector of hearing loss at the 6 audiometric frequencies
                  [250, 500, 1000, 2000, 4000, 6000] Hz.
    eq        Flag to provide equalization for the hearing loss to signal x:
                1 = no EQ has been provided, the function will add NAL-R
                2 = NAL-R EQ has already been added to the reference signal
    Level1    Optional input specifying level in dB SPL that corresponds to a
              signal RMS = 1. Default is 65 dB SPL if argument not provided.
    Returns:
    Combined  Quality estimate is the product of the nonlinear and linear terms
    Nonlin    Nonlinear quality component = (cepstral corr)^2 x seg BM coherence
    Linear    Linear quality component = std of spectrum and spectrum slope
    raw       Vector of raw values = [CepCorr, BMsync5, Dloud, Dslope]

    James M. Kates, 5 August 2013.
    Translated from MATLAB to Python by Gerardo Roa Dabike, October 2022.
    """

    # Auditory model for quality
    # Reference is no processing or NAL-R, impaired hearing
    xenv, xBM, yenv, yBM, xSL, ySL, fsamp = eb.EarModel(x, fx, y, fy, HL, eq, level1)

    # Envelope and long-term average spectral features
    # Smooth the envelope outputs: 125 Hz sub-sampling rate
    segsize = 16  # Averaging segment size in msec
    xdB = eb.env_smooth(xenv, segsize, fsamp)
    ydB = eb.env_smooth(yenv, segsize, fsamp)

    # Mel cepstrum correlation using smoothed envelopes
    # m1=ave of coefficients 2-6
    # xy=vector of coefficients 1-6
    thr = 2.5  # Silence threshold: sum across bands, dB above aud threshold
    addnoise = 0.0  # Additive noise in dB SL to condition cross-covariances
    cep_corr, xy = eb.melcor(xdB, ydB, thr, addnoise)

    # Linear changes in the log-term spectra
    # dloud  vector: [sum abs diff, std dev diff, max diff] spectra
    # dnorm  vector: [sum abs diff, std dev diff, max diff] norm spectra
    # dslope vector: [sum abs diff, std dev diff, max diff] slope
    dloud, dnorm, dslope = eb.spect_diff(xSL, ySL)

    # Temporal fine structure correlation measurements
    # Compute the time-frequency segment covariances
    segcov = 16  # Segment size for the covariance calculation
    sigcov, sigMSx, sigMSy = eb.bm_covary(xBM, yBM, segcov, fsamp)

    # Average signal segment cross-covariance
    # avecov=weighted ave of cross-covariances, using only data above threshold
    # syncov=ave cross-covariance with added IHC loss of synchronization at HF
    thr = 2.5  # Threshold in dB SL for including time-freq tile
    avecov, syncov = eb.ave_covary2(sigcov, sigMSx, thr)
    bm_sync5 = syncov[4]  # Ave segment coherence with IHC loss of sync

    # Extract and normalize the spectral features
    # Dloud:std
    d = dloud[1]  # Loudness difference std
    d = d / 2.5  # Scale the value
    d = 1.0 - d  # 1=perfect, 0=bad
    d = min(d, 1)
    d = max(d, 0)
    d_loud = d

    # Dslope:std
    d = dslope[1]  # Slope difference std
    d = 1.0 - d
    d = min(d, 1)
    d = max(d, 0)
    d_slope = d

    # Construct the models
    # Nonlinear model
    non_lin = (
        cep_corr**2
    ) * bm_sync5  # Combined envelope and temporal fine structure
    # Linear model
    linear = 0.579 * d_loud + 0.421 * d_slope  # Linear fit

    # Combined model
    combined = non_lin * linear  # Product of nonlinear x linear

    # Raw data
    raw = [cep_corr, bm_sync5, d_loud, d_slope]
    return combined, non_lin, linear, raw


def hasqi_v2_be(
    xl, xr, yl, yr, fs_signal, audiogram_l, audiogram_r, audiogram_cfs, level=100
) -> float:
    """Better ear HASQI.

    Calculates HASQI for left and right ear and selects the better result.

    Args:
        xl: left channel of reference signal
        xr: right channel of reference signal
        yl: left channel of processed signal
        yr: right channel of processed signal
        fs_signal: sampling rate for both signal
        audiogram_l: left ear audiogram
        audiogram_r: right ear audiogram
        audiogram_cfs: audiogram frequencies
        level: level in dB SPL corresponding to RMS=1

    Returns:
        float: beHASQI score

    Gerardo Roa Dabike, November 2022
    """

    # HASQI assumes the following audiogram frequencies:
    aud = [250, 500, 1000, 2000, 4000, 6000]

    # Adjust listener.audiogram_levels_l and _r to match the frequencies above
    hl_l = [
        audiogram_l[i] for i in range(len(audiogram_cfs)) if audiogram_cfs[i] in aud
    ]
    hl_r = [
        audiogram_r[i] for i in range(len(audiogram_cfs)) if audiogram_cfs[i] in aud
    ]

    score_l, _, _, _ = hasqi_v2(xl, fs_signal, yl, fs_signal, hl_l, eq=1, level1=level)
    score_r, _, _, _ = hasqi_v2(xr, fs_signal, yr, fs_signal, hl_r, eq=1, level1=level)

    return max(score_l, score_r)



import math
from typing import Optional, Tuple, Union

import numpy as np
import scipy
import scipy.signal


def firwin2(
    n: int,
    f: np.ndarray,
    a: np.ndarray,
    window: Optional[Union[str, float]] = None,
    antisymmetric: Optional[bool] = None,
) -> np.ndarray:
    """FIR filter design using the window method.
    Partial implementation of scipy firwin2 but using our own MATLAB-derived fir2.

    Args:
        n (int): The number of taps in the FIR filter.
        f (ndarray): The frequency sampling points. 0.0 to 1.0 with 1.0 being Nyquist.
        a (ndarray): The filter gains at the frequency sampling points.
        window (string or (string, float), optional): See scipy.firwin2 (default: (None))
        _antisymmetric (bool, optional): Unused but present to main compatability
            with scipy firwin2.
    Returns:
        np.ndarray:  The filter coefficients of the FIR filter, as a 1-D array of length n.
    """
    window_shape = None
    window_type: Union[str, float, None]
    window_param: Union[str, float]
    if isinstance(window, tuple):
        window_type, window_param = window if window is not None else (None, 0)
    else:
        window_type = window

    order = n - 1

    if window_type == "kaiser":
        window_shape = scipy.signal.kaiser(n, window_param)

    if window_shape is None:
        b, _ = fir2(order, f, a)
    else:
        b, _ = fir2(order, f, a, window_shape)

    return b


def fir2(
    nn: int, ff: np.ndarray, aa: np.ndarray, npt: Optional[int] = None
) -> Tuple[np.ndarray, int]:
    """FIR arbitrary shape filter design using the frequency sampling method.
    Translation of MATLAB fir2.

    Args:
        nn (int): Order
        ff (np.ndarray): Frequency breakpoints (0 < F < 1) where 1 is Nyquist rate.
                        First and last elements must be 0 and 1 respectively
        aa (np.ndarray): Magnitude breakpoints
        npt (int, optional): Number of points for freq response interpolation
            (default: max (smallest power of 2 greater than nn, 512))

    Returns:
        np.ndarray: nn + 1 filter coefficients
    """
    # Work with filter length instead of filter order
    nn += 1

    if npt is None:
        npt = 2.0 ** np.ceil(math.log(nn) / math.log(2)) if nn >= 1024 else 512
        wind = scipy.signal.hamming(nn)
    else:
        wind = npt
        npt = 2.0 ** np.ceil(math.log(nn) / math.log(2)) if nn >= 1024 else 512
    lap = int(np.fix(npt / 25))

    nbrk = max(len(ff), len(aa))

    ff[0] = 0
    ff[nbrk - 1] = 1

    H = np.zeros(npt + 1)
    nint: int = nbrk - 1
    df = np.diff(ff, n=1)

    npt += 1
    nb: int = 0
    H[0] = aa[0]

    for i in np.arange(nint):
        if df[i] == 0:
            nb = int(np.ceil(nb - lap / 2))
            ne = nb + lap - 1
        else:
            ne = int(np.fix(ff[i + 1] * npt)) - 1

        j = np.arange(nb, ne + 1)
        inc: Union[float, np.ndarray] = 0.0 if nb == ne else (j - nb) / (ne - nb)
        H[nb : (ne + 1)] = (inc * aa[i + 1]) + ((1 - inc) * aa[i])
        nb = ne + 1

    dt: float = 0.5 * (nn - 1)
    rad = -dt * 1j * math.pi * np.arange(0, npt) / (npt - 1)
    H = H * np.exp(rad)

    H = np.concatenate((H, H[npt - 2 : 0 : -1].conj()))
    ht = np.real(np.fft.ifft(H))

    b = ht[0:nn] * wind

    return b, 1


class NALR:
    def __init__(self, nfir: int, fs: int):
        """
        Args:
            nfir: Order of the NAL-R EQ filter and the matching delay
            fs: Sampling rate in Hz
        """
        self.nfir = nfir
        # Processing parameters
        self.fmax = 0.5 * fs

        # Audiometric frequencies
        self.aud = np.array([250, 500, 1000, 2000, 4000, 6000])

        # Design a flat filter having the same delay as the NAL-R filter
        self.delay = np.zeros(nfir + 1)
        self.delay[nfir // 2] = 1.0

    def hl_interp(self, hl: np.ndarray, cfs: np.ndarray):
        try:
            hl_interpf = scipy.interpolate.interp1d(cfs, hl)
        except ValueError:
            raise ValueError(
                "Hearing losses (hl) and center frequencies (cfs) don't match!"
            )
        return hl_interpf(self.aud)

    def build(
        self,
        hl: np.ndarray,
        cfs: np.ndarray = None,
    ):
        """
        Args:
            hl: hearing thresholds at [250, 500, 1000, 2000, 4000, 6000] Hz
            cfs: center frequencies of the hearing thresholds. If None, the default
                values are used.
        Returns:
            NAL-R FIR filter
            delay
        """
        if cfs is None:
            cfs = np.array([250, 500, 1000, 2000, 3000, 6000])

        hl = self.hl_interp(np.array(hl), np.array(cfs))
        mloss = np.max(hl)

        if mloss > 0:
            # Compute the NAL-R frequency response at the audiometric frequencies
            bias = np.array([-17, -8, 1, -1, -2, -2])
            t3 = hl[1] + hl[2] + hl[3]
            if t3 <= 180:
                xave = 0.05 * t3
            else:
                xave = 9.0 + 0.116 * (t3 - 180)
            gdB = xave + 0.31 * hl + bias
            gdB = np.clip(gdB, a_min=0, a_max=None)

            # Design the linear-phase FIR filter
            fv = np.append(
                np.append(0, self.aud), self.fmax
            )  # Frequency vector for the interpolation
            cfreq = (
                np.linspace(0, self.nfir, self.nfir + 1) / self.nfir
            )  # Uniform frequency spacing from 0 to 1
            gdBv = np.append(
                np.append(gdB[0], gdB), gdB[-1]
            )  # gdB vector for the interpolation
            interpf = scipy.interpolate.interp1d(fv, gdBv)
            gain = interpf(self.fmax * cfreq)
            glin = np.power(10, gain / 20.0)
            nalr = firwin2(self.nfir + 1, cfreq, glin)
        else:
            nalr = self.delay.copy()
        return nalr, self.delay

    def apply(self, nalr: np.ndarray, wav: np.ndarray):
        """
        Args:
            nalr: built NAL-R FIR filter
            wav: one dimensional wav signal

        Returns:
            amplified signal
        """
        return np.convolve(wav, nalr)



import numpy as np
from numba import jit
from scipy.signal import (
    butter,
    cheby2,
    convolve,
    correlate,
    firwin,
    group_delay,
    lfilter,
    resample_poly,
)

from clarity.enhancer.nalr import NALR


def EarModel(x, xsamp, y, ysamp, HL, itype, Level1):
    """
    Function to implement a cochlear model that includes the middle ear,
    auditory filter bank, OHC dynamic-range compression, and IHC attenuation.
    The inputs are the reference and processed signals that are to be
    compared. The reference x is at the reference intensity (e.g. 65 dB SPL
    or with NAL-R amplification) and has no other processing. The processed
    signal y is the hearing-aid output, and is assumed to have the same or
    greater group delay compared to the reference. The function outputs are
    the envelopes of the signals after OHC compression and IHC loss
    attenuation.

    Args:
    x        reference signal: should be adjusted to 65 dB SPL (itype=0 or 1)
               or to 65 dB SPL plus NAL-R gain (itype=2)
    xsamp    sampling rate for the reference signal, Hz
    y        processed signal (e.g. hearing-aid output) includes HA gain
    ysamp    sampling rate for the processed signal, Hz
    HL       audiogram giving the hearing loss in dB at six audiometric
               frequencies: [250, 500, 1000, 2000, 4000, 6000] Hz
    itype    purpose for the calculation:
             0=intelligibility: reference is nornal hearing and must not
               include NAL-R EQ
             1=quality: reference does not include NAL-R EQ
             2=quality: reference already has NAL-R EQ applied
    Level1   level calibration: signal RMS=1 corresponds to Level1 dB SPL

    Returns:
    xdB      envelope for the reference in each band
    xBM      BM motion for the reference in each band
    ydB      envelope for the processed signal in each band
    yBM      BM motion for the processed signal in each band
    xSL      compressed RMS average reference in each band converted to dB SL
    ySL      compressed RMS average output in each band converted to dB SL
    fsamp    sampling rate in Hz for the model outputs

    James M. Kates, 27 October 2011.
    BM motion added 30 Dec 2011.
    Revised 19 June 2012.
    Remove match of reference RMS level to processed 29 August 2012.
    IHC adaptation added 1 October 2012.
    BM envelope coverted to dB SL, 2 Oct 2012.
    Filterbank group delay corrected, 14 Dec 2012.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    Updated by Gerardo Roa Dabike, September 2022.
    """

    # Processing parameters
    # OHC and IHC parameters for the hearing loss
    # Auditory filter center frequencies span 80 to 8000 Hz.
    nchan = 32  # Use 32 auditory frequency bands
    mdelay = 1  # Compensate for the gammatone group delay
    cfreq = CenterFreq(nchan)  # Center frequencies on an ERB scale

    # Cochlear model parameters for the processed signal
    attnOHCy, BWminy, lowkneey, CRy, attnIHCy = LossParameters(HL, cfreq)

    # The cochlear model parameters for the reference are the same as for the hearing
    # loss if calculating quality, but are for normal hearing if calculating intelligibility.
    if itype == 0:
        HLx = [0] * len(HL)
    else:
        HLx = HL
    [attnOHCx, BWminx, lowkneex, CRx, attnIHCx] = LossParameters(HLx, cfreq)

    # Parameters for the control filter bank
    HLmax = [100, 100, 100, 100, 100, 100]
    shift = 0.02  # Basal shift of 0.02 of the basilar membrane length
    cfreq1 = CenterFreq(nchan, shift)  # Center frequencies for the control
    _, BW1, _, _, _ = LossParameters(HLmax, cfreq1)
    # Maximum BW for the control

    # Input signal adjustments
    # Convert the signals to 24 kHz sampling rate.
    # Using 24 kHz guarantees that all of the cochlear filters have the same shape
    # independent of the incoming signal sampling rates
    x24, _ = Resamp24kHz(x, xsamp)
    y24, fsamp = Resamp24kHz(y, ysamp)

    # Check file sizes
    nxy = min(len(x24), len(y24))
    x24 = x24[:nxy]
    y24 = y24[:nxy]

    # Bulk broadband signal alignment
    x24, y24 = InputAlign(x24, y24)
    nsamp = len(x24)

    # For HASQI, here add NAL-R equalization if the quality reference doesn't already have it.
    if itype == 1:
        nfir = 140  # Length in samples of the FIR NAL-R EQ filter (24-kHz rate)
        enhancer = NALR(nfir, fsamp)
        aud = [250, 500, 1000, 2000, 4000, 6000]
        nalr_fir, _ = enhancer.build(HL, aud)
        x24 = convolve(x24, nalr_fir)  # Apply the NAL-R filter
        x24 = x24[nfir : nfir + nsamp]

    # Cochlear model
    # Middle ear
    xmid = MiddleEar(x24, fsamp)
    ymid = MiddleEar(y24, fsamp)

    # Initialize storage
    # Reference and processed envelopes and BM motion
    xdB = np.zeros((nchan, nsamp))
    ydB = np.zeros((nchan, nsamp))

    # Reference and processed average spectral values
    xave = np.zeros(nchan)
    yave = np.zeros(nchan)  # Processed
    xcave = np.zeros(nchan)  # Reference control
    ycave = np.zeros(nchan)  # Processed control

    # Filter bandwidths adjusted for intensity
    BWx = np.zeros(nchan)
    BWy = np.zeros(nchan)

    xb = np.zeros((nchan, nsamp))
    yb = np.zeros((nchan, nsamp))

    # Loop over each filter in the auditory filter bank
    for n in range(nchan):
        # Control signal envelopes for the reference and processed signals
        xcontrol, _, ycontrol, _ = GammatoneBM(
            xmid, BW1[n], ymid, BW1[n], fsamp, cfreq1[n]
        )

        # Adjust the auditory filter bandwidths for the average signal level
        BWx[n] = BWadjust(xcontrol, BWminx[n], BW1[n], Level1)  # Reference
        BWy[n] = BWadjust(ycontrol, BWminy[n], BW1[n], Level1)  # Processed

        # Envelopes and BM motion of the reference and processed signals
        xenv, xbm, yenv, ybm = GammatoneBM(xmid, BWx[n], ymid, BWy[n], fsamp, cfreq[n])

        # RMS levels of the ref and output envelopes for linear metric
        xave[n] = np.sqrt(np.mean(xenv**2))
        yave[n] = np.sqrt(np.mean(yenv**2))
        xcave[n] = np.sqrt(np.mean(xcontrol**2))
        ycave[n] = np.sqrt(np.mean(ycontrol**2))

        # Cochlear compression for the signal envelopes and BM motion
        xc, xb[n] = EnvCompressBM(
            xenv, xbm, xcontrol, attnOHCx[n], lowkneex[n], CRx[n], fsamp, Level1
        )
        yc, yb[n] = EnvCompressBM(
            yenv, ybm, ycontrol, attnOHCy[n], lowkneey[n], CRy[n], fsamp, Level1
        )

        # Correct for the delay between the reference and output
        yc = EnvAlign(xc, yc)  # Align processed envelope to reference
        yb[n] = EnvAlign(xb[n], yb[n])  # Align processed BM motion to reference

        # Convert the compressed envelopes and BM vibration envelopes to dB SPL
        xc, xb[n] = EnvSL(xc, xb[n], attnIHCx[n], Level1)
        yc, yb[n] = EnvSL(yc, yb[n], attnIHCy[n], Level1)

        # Apply the IHC rapid and short-term adaptation
        delta = 2  # Amount of overshoot
        xdB[n], xb[n] = IHCadapt(xc, xb[n], delta, fsamp)
        ydB[n], yb[n] = IHCadapt(yc, yb[n], delta, fsamp)

    # Additive noise level to give the auditory threshold
    IHCthr = -10  # Additive noise level, dB re: auditory threshold
    xBM = BMaddnoise(xb, IHCthr, Level1)
    yBM = BMaddnoise(yb, IHCthr, Level1)

    # Correct for the gammatone filterbank interchannel group delay.
    if mdelay > 0:
        xdB = GroupDelayComp(xdB, BWx, cfreq, fsamp)
        ydB = GroupDelayComp(ydB, BWx, cfreq, fsamp)
        xBM = GroupDelayComp(xBM, BWx, cfreq, fsamp)
        yBM = GroupDelayComp(yBM, BWx, cfreq, fsamp)

    # Convert average gammatone outputs to dB SPL
    xSL = aveSL(xave, xcave, attnOHCx, lowkneex, CRx, attnIHCx, Level1)
    ySL = aveSL(yave, ycave, attnOHCy, lowkneey, CRy, attnIHCy, Level1)

    return xdB, xBM, ydB, yBM, xSL, ySL, fsamp


def CenterFreq(nchan, shift=None):
    """
    Function to compute the ERB frequency spacing for the gammatone
    filter bank. The equation comes from Malcolm Slaney (1993).

    Calling variables
    nchan		number of filters in the filter bank
    shift     optional frequency shift of the filter bank specified as a
              fractional shift in distance along the BM. A positive shift
              is an increase in frequency (basal shift), and negative is
              a decrease in frequency (apical shift). The total length of
              the BM is normalized to 1. The frequency-to-distance map is
              from D.D. Greenwood (1990), JASA 87, 2592-2605, Eq (1).

    James M. Kates, 25 January 2007.
    Frequency shift added 22 August 2008.
    Lower and upper frequencies fixed at 80 and 8000 Hz, 19 June 2012.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    lowFreq = 80
    highFreq = 8000

    # Moore and Glasberg ERB values
    EarQ = 9.26449
    minBW = 24.7

    # In the Matlab code, the loop below never evaluates
    # (but the current code was trained with this bug)
    shift = None  # This is to keep consistency with MATLAB code
    if shift is not None:
        k = 1
        A = 165.4
        a = 2.1  # shift specified as a fraction of the total length
        # Locations of the low and high frequencies on the BM between 0 and 1
        xLow = (1 / a) * np.log10(k + (lowFreq / A))
        xHigh = (1 / a) * np.log10(k + (highFreq / A))
        # Shift the locations
        xLow = xLow * (1 + shift)
        xHigh = xHigh * (1 + shift)
        # Compute the new frequency range
        lowFreq = A * (10 ** (a * xLow) - k)
        highFreq = A * (10 ** (a * xHigh) - k)

    # All of the following expressions are derived in Apple TR #35,
    # "An Efficient Implementation of the Patterson-Holdsworth Cochlear
    # Filter Bank" by Malcolm Slaney.
    cf = -(EarQ * minBW) + np.exp(
        np.arange(1, nchan)
        * (-np.log(highFreq + EarQ * minBW) + np.log(lowFreq + EarQ * minBW))
        / (nchan - 1)
    ) * (highFreq + EarQ * minBW)
    cf = np.insert(cf, 0, highFreq)  # Last center frequency is set to highFreq
    cf = np.flip(cf)
    return cf


def LossParameters(HL, cfreq):
    """
    Function to apportion the hearing loss to the outer hair cells (OHC)
    and the inner hair cells (IHC) and to increase the bandwidth of the
    cochlear filters in proportion to the OHC fraction of the total loss.

    Calling variables:
    HL		hearing loss at the 6 audiometric frequencies
    cfreq		array containing the center frequencies of the gammatone filters
                arranged from low to high

    Returns:
    attnOHC	attenuation in dB for the OHC gammatone filters
    BW		OHC filter bandwidth expressed in terms of normal
    lowknee	Lower kneepoint for the low-level linear amplification
    CR		Ranges from 1.4:1 at 150 Hz to 3.5:1 at 8 kHz for normal
                hearing. Reduced in proportion to the OHC loss to 1:1.
    attnIHC	attenuation in dB for the input to the IHC synapse

    James M. Kates, 25 January 2007.
    Version for loss in dB and match of OHC loss to CR, 9 March 2007.
    Low-frequency extent changed to 80 Hz, 27 Oct 2011.
    Lower kneepoint set to 30 dB, 19 June 2012.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Audiometric frequencies in Hz
    aud = [250, 500, 1000, 2000, 4000, 6000]

    # Interpolation to give the loss at the gammatone center frequencies
    # Use linear interpolation in dB. The interpolation assumes that
    # cfreq[1] < aud[1] and cfreq[nfilt] > aud[6]
    nfilt = len(cfreq)
    fv = np.insert(aud, [0, len(aud)], [cfreq[0], cfreq[-1]])

    # Interpolated gain in dB
    loss = np.interp(cfreq, fv, np.insert(HL, [0, len(HL)], [HL[0], HL[-1]]))
    loss = np.maximum(loss, 0)
    # Make sure there are no negative losses

    # Compression ratio changes linearly with ERB rate from 1.25:1 in the 80-Hz
    # frequency band to 3.5:1 in the 8-kHz frequency band
    CR = 1.25 + 2.25 * np.arange(nfilt) / (nfilt - 1)

    # Maximum OHC sensitivity loss depends on the compression ratio.
    # The compression I/O curves assume linear below 30 and above 100 dB SPL in normal ears.
    maxOHC = 70 * (1 - (1 / CR))  # HC loss that results in 1:1 compression
    thrOHC = 1.25 * maxOHC  # Loss threshold for adjusting the OHC parameters

    # Apportion the loss in dB to the outer and inner hair cells based on the data of
    # Moore et al (1999), JASA 106, 2761-2778.

    # Reduce the CR towards 1:1 in proportion to the OHC loss.
    attnOHC = 0.8 * np.copy(loss)
    attnIHC = 0.2 * np.copy(loss)

    attnOHC[loss >= thrOHC] = 0.8 * thrOHC[loss >= thrOHC]
    attnIHC[loss >= thrOHC] = 0.2 * thrOHC[loss >= thrOHC] + (
        loss[loss >= thrOHC] - thrOHC[loss >= thrOHC]
    )

    # Adjust the OHC bandwidth in proportion to the OHC loss
    BW = np.ones(nfilt)
    BW = BW + (attnOHC / 50.0) + 2.0 * (attnOHC / 50.0) ** 6

    # Compute the compression lower kneepoint and compression ratio
    lowknee = attnOHC + 30
    upamp = 30 + 70 / CR  # Output level for an input of 100 dB SPL

    CR = (100 - lowknee) / (upamp + attnOHC - lowknee)  # OHC loss Compression ratio

    return attnOHC, BW, lowknee, CR, attnIHC


def Resamp24kHz(x, fsampx):
    """
    Function to resample the input signal at 24 kHz. The input sampling rate
    is rounded to the nearest kHz to comput the sampling rate conversion
    ratio.

    Calling variables:
    x         input signal
    fsampx    sampling rate for the input in Hz

    Returned argument:
    y         signal resampled at 24 kHz
    fsamp     output sampling rate in Kz

    James M. Kates, 20 June 2011.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """

    # Sampling rate information
    fsamp = 24000
    fy = round(fsamp / 1000)  # output rate to nearest kHz
    fx = round(fsampx / 1000)

    # Resample the signal
    if fx == fy:
        # No resampling performed if the rates match
        y = x
    elif fx < fy:
        # Resample for the input rate lower than the output
        y = resample_poly(x, fy, fx)

        # Match the RMS level of the resampled signal to that of the input
        xRMS = np.sqrt(np.mean(x**2))
        yRMS = np.sqrt(np.mean(y**2))
        y = (xRMS / yRMS) * y

    else:
        # Resample for the input rate higher than the output
        y = resample_poly(x, fy, fx)

        # Reduce the input signal bandwidth to 21 kHz (-10.5 to +10.5 kHz)
        # The power equalization is designed to match the signal intensities
        # over the frequency range spanned by the gammatone filter bank.
        # Chebyshev Type 2 LP
        order = 7
        atten = 30  # sidelobe attenuation in dB
        fcutx = 21 / fx
        bx, ax = cheby2(order, atten, fcutx)
        xfilt = lfilter(bx, ax, x, axis=0)

        # Reduce the resampled signal bandwisth to 21 kHz (-10.5 to +10.5 kHz)
        fcuty = 21 / fy
        by, ay = cheby2(order, atten, fcuty)
        yfilt = lfilter(by, ay, y, axis=0)

        # Compute the input and output RMS levels within the 21 kHz bandwidth and
        # match the output to the input
        xRMS = np.sqrt(np.mean(xfilt**2))
        yRMS = np.sqrt(np.mean(yfilt**2))
        y = (xRMS / yRMS) * y

    return y, fsamp


def InputAlign(x, y):
    """
    Function to provide approximate temporal alignment of the reference and
    processed output signals. Leading and trailing zeros are then pruned.
    The function assumes that the two sequences have the same sampling rate:
    call eb_Resamp24kHz for each sequence first, then call this function to
    align the signals.

    Calling variables:
    x       input reference sequence
    y       hearing-aid output sequence

    Returns:
    xp   pruned and shifted reference
    yp   pruned and shifted hearing-aid output

    James M. Kates, 12 July 2011.
    Match the length of the processed output to the reference for the
    purposes of computing the cross-covariance
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """

    # Match the length of the processed output to the reference for the purposes
    # of computing the cross-covariance
    nx = len(x)
    ny = len(y)
    nsamp = min(nx, ny)

    # Determine the delay of the output relative to the reference
    xy = correlate(
        x[:nsamp] - np.mean(x[:nsamp]), y[:nsamp] - np.mean(y[:nsamp]), "full"
    )  # Matlab code uses xcov thus the subtraction of mean
    index = np.argmax(np.abs(xy))
    delay = nsamp - index - 1

    # Back up 2 msec to allow for dispersion
    fsamp = 24000  # Cochlear model input sampling rate in Hz
    delay = round(delay - 2 * fsamp / 1000)  # Back up 2 ms

    # Align the output with the reference allowing for the dispersion
    if delay > 0:
        # Output delayed relative to the reference
        y = np.concatenate((y[delay:ny], np.zeros(delay)))
    else:
        # Output advanced relative to the reference
        y = np.concatenate((np.zeros(-delay), y[: ny + delay]))

    # Find the start and end of the noiseless reference sequence
    xabs = np.abs(x)
    xmax = np.max(xabs)
    xthr = 0.001 * xmax  # Zero detection threshold

    above_threshold = np.where(xabs > xthr)[0]
    nx0 = above_threshold[0]
    nx1 = above_threshold[-1]

    # Prune the sequences to remove the leading and trailing zeros
    nx1 = min(nx1, ny)
    xp = x[nx0 : nx1 + 1]
    yp = y[nx0 : nx1 + 1]

    return xp, yp


def MiddleEar(x, fsamp):
    """
    Function to design the middle ear filters and process the input
    through the cascade of filters. The middle ear model is a 2-pole HP
    filter at 350 Hz in series with a 1-pole LP filter at 5000 Hz. The
    result is a rough approximation to the equal-loudness contour at
    threshold.

    Calling variables:
        x		input signal
        fsamp	sampling rate in Hz

    Function output:
        xout	filtered output

    James M. Kates, 18 January 2007.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """

    # Design the 1-pole Butterworth LP using the bilinear transformation
    bLP, aLP = butter(1, 5000 / (0.5 * fsamp))

    # LP filter the input
    y = lfilter(bLP, aLP, x)

    # Design the 2-pole Butterworth HP using the bilinear transformation
    bHP, aHP = butter(2, 350 / (0.5 * fsamp), "high")

    # HP fitler the signal
    xout = lfilter(bHP, aHP, y)

    return xout


def GammatoneBM(x, BWx, y, BWy, fs, cf):
    """
    4th-order gammatone auditory filter. This implementation is based
    on the c program published on-line by Ning Ma, U. Sheffield, UK,
    that gives an implementation of the Martin Cooke (1991) filters:
    an impulse-invariant transformation of the gammatone filter. The
    signal is demodulated down to baseband using a complex exponential,
    and then passed through a cascade of four one-pole low-pass filters.

    This version filters two signals that have the same sampling rate and the
    same gammatone filter center frequencies. The lengths of the two signals
    should match; if they don't, the signals are truncated to the shorter of
    the two lengths.

    Calling variables:
    x			first sequence to be filtered
    BWx	    bandwidth for x relative to that of a normal ear
    y			second sequence to be filtered
    BWy	    bandwidth for x relative to that of a normal ear
    fs		sampling rate in Hz
    cf		filter center frequency in Hz

    Returns:
    envx      filter envelope output (modulated down to baseband) 1st signal
    BMx       BM motion for the first signal
    envy      filter envelope output (modulated down to baseband) 2nd signal
    BMy       BM motion for the second signal
    James M. Kates, 8 Jan 2007.
    Vectorized version for efficient MATLAB execution, 4 February 2007.
    Cosine and sine generation, 29 June 2011.
    Output sine and cosine sequences, 19 June 2012.
    Cosine/sine loop speed increased, 9 August 2013.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Filter ERB from Moore and Glasberg (1983)
    earQ = 9.26449
    minBW = 24.7
    ERB = minBW + (cf / earQ)

    # Check the lengths of the two signals
    nx = len(x)
    ny = len(y)
    nsamp = min(nx, ny)
    x = x[:nsamp]
    y = y[:nsamp]

    # Filter the first signal
    # Initialize the filter coefficients
    tpt = 2 * np.pi / fs
    tptBW = BWx * tpt * ERB * 1.019
    a = np.exp(-tptBW)
    a1 = 4.0 * a
    a2 = -6.0 * a * a
    a3 = 4.0 * a * a * a
    a4 = -a * a * a * a
    a5 = 4.0 * a * a
    gain = 2.0 * (1 - a1 - a2 - a3 - a4) / (1 + a1 + a5)

    # Initialize the complex demodulation
    npts = len(x)
    sincf, coscf = GammatoneBW_demodulation(
        npts, tpt, cf, np.zeros(npts), np.zeros(npts)
    )

    # Filter the real and imaginary parts of the signal
    ureal = lfilter([1, a1, a5], [1, -a1, -a2, -a3, -a4], x * coscf)
    uimag = lfilter([1, a1, a5], [1, -a1, -a2, -a3, -a4], x * sincf)

    # Extract the BM velocity and the envelope
    BMx = gain * (ureal * coscf + uimag * sincf)
    envx = gain * np.sqrt(ureal * ureal + uimag * uimag)

    # Filter the second signal using the existing cosine and sine sequences
    tptBW = BWy * tpt * ERB * 1.019
    a = np.exp(-tptBW)
    a1 = 4.0 * a
    a2 = -6.0 * a * a
    a3 = 4.0 * a * a * a
    a4 = -a * a * a * a
    a5 = 4.0 * a * a
    gain = 2.0 * (1 - a1 - a2 - a3 - a4) / (1 + a1 + a5)

    # Filter the real and imaginary parts of the signal
    ureal = lfilter([1, a1, a5], [1, -a1, -a2, -a3, -a4], y * coscf)
    uimag = lfilter([1, a1, a5], [1, -a1, -a2, -a3, -a4], y * sincf)

    # Extract the BM velocity and the envelope
    BMy = gain * (ureal * coscf + uimag * sincf)
    envy = gain * np.sqrt(ureal * ureal + uimag * uimag)

    return envx, BMx, envy, BMy


@jit(nopython=True)
def GammatoneBW_demodulation(npts, tpt, cf, coscf, sincf):
    cn = np.cos(tpt * cf)
    sn = np.sin(tpt * cf)
    cold = 1
    sold = 0
    coscf[0] = cold
    sincf[0] = sold
    for n in range(1, npts):
        arg = cold * cn + sold * sn
        sold = sold * cn - cold * sn
        cold = arg
        coscf[n] = cold
        sincf[n] = sold

    return sincf, coscf


def BWadjust(control, BWmin, BWmax, Level1):
    """
    Function to compute the increase in auditory filter bandwidth in response
    to high signal levels.

    Args:
    control     envelope output in the control filter band
    BWmin       auditory filter bandwidth computed for the loss (or NH)
    BWmax       auditory filter bandwidth at maximum OHC damage
    Level1      RMS=1 corresponds to Level1 dB SPL

    Returned value:
    BW          filter bandwidth increased for high signal levels

    James M. Kates, 21 June 2011.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """

    # Compute the control signal level
    cRMS = np.sqrt(np.mean(control**2))
    cdB = 20 * np.log10(cRMS) + Level1

    # Adjust the auditory filter bandwidth
    if cdB < 50:
        # No BW adjustment for a signal below 50 dB SPL
        BW = BWmin
    elif cdB > 100:
        # Maximum BW if signal is above 100 dB SPL
        BW = BWmax
    else:
        # Linear interpolation between BW at 50 dB and max BW at 100 dB SPL
        BW = BWmin + ((cdB - 50) / 50) * (BWmax - BWmin)

    return BW


def EnvCompressBM(envsig, bm, control, attnOHC, thrLow, CR, fsamp, Level1):
    """
    Function to compute the cochlear compression in one auditory filter
    band. The gain is linear below the lower threshold, compressive with
    a compression ratio of CR:1 between the lower and upper thresholds,
    and reverts to linear above the upper threshold. The compressor
    assumes that auditory thresold is 0 dB SPL.

    Calling variables:
    envsig	analytic signal envelope (magnitude) returned by the
                gammatone filter bank
    bm        BM motion output by the filter bank
    control	analytic control envelope returned by the wide control
                path filter bank
    attnOHC	OHC attenuation at the input to the compressor
    thrLow	kneepoint for the low-level linear amplification
    CR		compression ratio
    fsamp		sampling rate in Hz
    Level1	dB reference level: a signal having an RMS value of 1 is
                assigned to Level1 dB SPL.

    Function outputs:
    y			compressed version of the signal envelope
    b         compressed version of the BM motion

    James M. Kates, 19 January 2007.
    LP filter added 15 Feb 2007 (Ref: Zhang et al., 2001)
    Version to compress the envelope, 20 Feb 2007.
    Change in the OHC I/O function, 9 March 2007.
    Two-tone suppression added 22 August 2008.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Initialize the compression parameters
    thrHigh = 100

    # Convert the control envelope to dB SPL
    small = 1e-30
    logenv = np.maximum(control, small)
    logenv = Level1 + 20 * np.log10(logenv)
    logenv = np.minimum(logenv, thrHigh)  # Clip signal levels above the upper threshold
    logenv = np.maximum(logenv, thrLow)  # Clip signal at the lower threshold

    # Compute the compression gain in dB
    gain = -attnOHC - (logenv - thrLow) * (1 - (1 / CR))

    # Convert the gain to linear and apply a LP filter to give a 0.2 ms delay
    gain = 10 ** (gain / 20)
    flp = 800
    b, a = butter(1, flp / (0.5 * fsamp))
    gain = lfilter(b, a, gain)

    # Apply the gain to the signals
    y = gain * envsig
    b = gain * bm

    return y, b


def EnvAlign(x, y):
    """
    Function to align the envelope of the processed signal to that of the
    reference signal.

    Args:
    x      envelope or BM motion of the reference signal
    y      envelope or BM motion of the output signal

    Returns:
    y      shifted output envelope to match the input

    James M. Kates, 28 October 2011.
    Absolute value of the cross-correlation peak removed, 22 June 2012.
    Cross-correlation range reduced, 13 August 2013.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """

    # The MATLAB code limits the range of lags to search (to 100 ms) to save
    # computation time - no such option exists in numpy, but the code below limits the delay
    # to the same range as in Matlab, for consistent results
    fsamp = 24000
    corr_range = 100  # Range in msec for the correlation
    lags = round(0.001 * corr_range * fsamp)  # Range in samples
    npts = len(x)
    lags = min(lags, npts)

    xy = correlate(x, y, "full")
    location = np.argmax(xy[npts - lags : npts + lags])  # Limit the range in which
    delay = lags - location - 1

    # Time shift the output sequence
    if delay > 0:
        # Output delayed relative to the reference
        y = np.concatenate((y[delay:npts], np.zeros(delay)))
    elif delay < 0:
        # Output advanced relative to the reference
        y = np.concatenate((np.zeros(-delay), y[: npts + delay]))

    return y


def EnvSL(env, bm, attnIHC, Level1):
    """
    Function to convert the compressed envelope returned by
    cochlea_envcomp to dB SL.

    Args
    env			linear envelope after compression
    bm            linear basilar membrane vibration after compression
    attnIHC		IHC attenuation at the input to the synapse
    Level1		level in dB SPL corresponding to 1 RMS

    Return
    y				envelope in dB SL
    b             BM vibration with envelope converted to dB SL

    James M. Kates, 20 Feb 07.
    IHC attenuation added 9 March 2007.
    Basilar membrane vibration conversion added 2 October 2012.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Convert the envelope to dB SL
    small = 1e-30
    y = Level1 - attnIHC + 20 * np.log10(env + small)
    y = np.maximum(y, 0)

    # Convert the linear BM motion to have a dB SL envelope
    gain = (y + small) / (env + small)
    b = gain * bm

    return y, b


@jit(nopython=True)
def IHCadapt(xdB, xBM, delta, fsamp):
    """
    Function to provide inner hair cell (IHC) adaptation. The adaptation is
    based on an equivalent RC circuit model, and the derivatives are mapped
    into 1st-order backward differences. Rapid and short-term adaptation are
    provided. The input is the signal envelope in dB SL, with IHC attenuation
    already applied to the envelope. The outputs are the envelope in dB SL
    with adaptation providing overshoot of the long-term output level, and
    the BM motion is multiplied by a gain vs. time function that reproduces
    the adaptation. IHC attenuation and additive noise for the equivalent
    auditory threshold are provided by a subsequent call to eb_BMatten.

    Calling variables:
    xdB      signal envelope in one frequency band in dB SL
             contains OHC compression and IHC attenuation
    xBM      basilar membrane vibration with OHC compression but no IHC atten
    delta    overshoot factor = delta x steady-state
    fsamp    sampling rate in Hz

    Returns:
    ydB      envelope in dB SL with IHC adaptation
    yBM      BM motion multiplied by the IHC adaptation gain function

    James M. Kates, 1 October 2012.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Test the amount of overshoot
    dsmall = 1.0001
    delta = max(delta, dsmall)

    # Initialize adaptation time constants
    tau1 = 2  # Rapid adaptation in msec
    tau2 = 60  # Short-term adaptation in msec
    tau1 = 0.001 * tau1  # Convert to seconds
    tau2 = 0.001 * tau2

    # Equivalent circuit parameters
    T = 1 / fsamp
    R1 = 1 / delta
    R2 = 0.5 * (1 - R1)
    R3 = R2
    C1 = tau1 * (R1 + R2) / (R1 * R2)
    C2 = tau2 / ((R1 + R2) * R3)

    # Intermediate values used for the voltage update matrix inversion
    a11 = R1 + R2 + R1 * R2 * (C1 / T)
    a12 = -R1
    a21 = -R3
    a22 = R2 + R3 + R2 * R3 * (C2 / T)
    denom = 1 / (a11 * a22 - a21 * a12)

    # Additional intermediate values
    R1inv = 1 / R1
    R12C1 = R1 * R2 * (C1 / T)
    R23C2 = R2 * R3 * (C2 / T)

    # Initalize the outputs and state of the equivalent circuit
    nsamp = len(xdB)
    gain = np.ones_like(xdB)  # Gain vector to apply to the BM motion, default is 1
    ydB = np.zeros_like(xdB)
    V1 = 0
    V2 = 0
    small = 1e-30

    # Loop to process the envelope signal
    # The gain asymptote is 1 for an input envelope of 0 dB SPL
    for n in range(nsamp):
        V0 = xdB[n]
        b1 = V0 * R2 + R12C1 * V1
        b2 = R23C2 * V2
        V1 = denom * (a22 * b1 - a12 * b2)
        V2 = denom * (-a21 * b1 + a11 * b2)
        out = (V0 - V1) * R1inv
        ydB[n] = out

    ydB = np.maximum(ydB, 0)
    gain = (ydB + small) / (xdB + small)

    yBM = gain * xBM

    return ydB, yBM


def BMaddnoise(x, thr, Level1):
    """
    Function to apply the IHC attenuation to the BM motion and to add a
    low-level Gaussian noise to give the auditory threshold.

    Args:
    x         BM motion to be attenuated
    thr       additive noise level in dB re:auditory threshold
    Level1    an input having RMS=1 corresponds to Leve1 dB SPL

    Returns:
    y         attenuated signal with threhsold noise added

    James M. Kates, 19 June 2012.
    Just additive noise, 2 Oct 2012.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    gn = 10 ** ((thr - Level1) / 20)  # Linear gain for the noise

    # rng = np.random.default_rng()
    noise = gn * np.random.standard_normal(x.shape)  # Gaussian RMS=1, then attenuated
    y = x + noise

    return y


def GroupDelayComp(xenv, BW, cfreq, fsamp):
    """
    Function to compensate for the group delay of the gammatone filter bank.
    The group delay is computed for each filter at its center frequency. The
    firing rate output of the IHC model is then adjusted so that all outputs
    have the same group delay.

    Calling variables:
        xenv (np.ndarray): matrix of signal envelopes or BM motion
        BW (): gammatone filter bandwidths adjusted for loss
        cfreq (): center frequencies of the bands
        fsamp (): sampling rate for the input signal in Hz (e.g. 24,000 Hz)

    Returns:
        yenv    envelopes or BM motion compensated for the group delay

    James M. Kates, 28 October 2011.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """
    # Processing parameters
    nchan = len(BW)

    # Filter ERB from Moore and Glasberg (1983)
    earQ = 9.26449
    minBW = 24.7
    ERB = minBW + (cfreq / earQ)

    # Initialize the gamatone filter coefficients
    tpt = 2 * np.pi / fsamp
    tptBW = tpt * 1.019 * BW * ERB
    a = np.exp(-tptBW)
    a1 = 4.0 * a
    a2 = -6.0 * a * a
    a3 = 4.0 * a * a * a
    a4 = -a * a * a * a
    a5 = 4.0 * a * a

    # Compute the group delay in samples at fsamp for each filter
    gd = np.zeros(nchan)
    for n in range(nchan):
        _, gd[n] = group_delay(
            ([1, a1[n], a5[n]], [1, -a1[n], -a2[n], -a3[n], -a4[n]]), 1
        )
    gd = np.round(gd).astype("int")  # convert to integer samples

    # Compute the delay correlation
    gmin = np.min(gd)
    gd = gd - gmin  # Remove the minimum delay from all the over values
    gmax = np.max(gd)
    correct = gmax - gd  # Samples delay needed to add to give alignment

    # Add delay correction to each frequency band
    yenv = np.zeros(xenv.shape)
    for n in range(nchan):
        r = xenv[n]
        npts = len(r)
        yenv[n] = np.concatenate((np.zeros(correct[n]), r[: npts - correct[n]]))

    return yenv


def aveSL(env, control, attnOHC, thrLow, CR, attnIHC, Level1):
    """
    Function to covert the RMS average output of the gammatone filter bank
    into dB SL. The gain is linear below the lower threshold, compressive
    with a compression ratio of CR:1 between the lower and upper thresholds,
    and reverts to linear above the upper threshold. The compressor
    assumes that auditory thresold is 0 dB SPL.

    Calling variables:
    env		analytic signal envelope (magnitude) returned by the
                gammatone filter bank, RMS average level
    control   control signal envelope
    attnOHC	OHC attenuation at the input to the compressor
    thrLow	kneepoint for the low-level linear amplification
    CR		compression ratio
    attnIHC	IHC attenuation at the input to the synapse
    Level1	dB reference level: a signal having an RMS value of 1 is
                assigned to Level1 dB SPL.

    Function output:
    xdB		compressed output in dB above the impaired threshold

    James M. Kates, 6 August 2007.
    Version for two-tone suppression, 29 August 2008.
    Translated from MATLAB to Python by Zuzanna Podwinska, March 2022.
    """

    # Initialize the compression parameters
    thrHigh = 100  # Upper compression threshold

    # Convert the control to dB SPL
    small = 1e-30
    logenv = np.maximum(control, small)
    logenv = Level1 + 20 * np.log10(logenv)
    logenv = np.minimum(logenv, thrHigh)
    logenv = np.maximum(logenv, thrLow)

    # Compute compression gain in dB
    gain = -attnOHC - (logenv - thrLow) * (1 - (1 / CR))

    # Convert the signal envelope to dB SPL
    logenv = np.maximum(env, small)
    logenv = Level1 + 20 * np.log10(logenv)
    logenv = np.maximum(logenv, 0)
    xdB = logenv + gain - attnIHC
    xdB = np.maximum(xdB, 0)

    return xdB


def env_smooth(env, segsize, fsamp):
    """
    Function to smooth the envelope returned by the cochlear model. The
    envelope is divided into segments having a 50% overlap. Each segment is
    windowed, summed, and divided by the window sum to produce the average.
    A raised cosine window is used. The envelope sub-sampling frequency is
    2*(1000/segsize).

    Arguments:
        xenv: matrix of envelopes in each of the auditory bands
        segsize: averaging segment size in msec
        fsamp: input envelope sampling rate in Hz

    Returns:
        smooth: matrix of subsampled windowed averages in each band

    James M. Kates, 26 January 2007.
    Final half segment added 27 August 2012.

    Translated from MATLAB to Python by Gerardo Roa Dabike, September 2022.
    """

    # Compute the window
    nwin = int(np.around(segsize * (0.001 * fsamp)))  # Segment size in samples
    test = nwin - 2 * np.floor(nwin / 2)  # 0=even, 1=odd
    if test > 0:
        # Force window length to be even
        nwin = nwin + 1
    window = np.hanning(nwin)  # Raised cosine von Hann window
    wsum = np.sum(window)  # Sum for normalization

    #  The first segment has a half window
    nhalf = int(nwin / 2)
    halfwindow = window[nhalf:nwin]
    halfsum = np.sum(halfwindow)

    # Number of segments and assign the matrix storage
    nchan = np.size(env, 0)
    npts = np.size(env, 1)
    nseg = int(1 + np.floor(npts / nwin) + np.floor((npts - nwin / 2) / nwin))
    smooth = np.zeros((nchan, nseg))

    #  Loop to compute the envelope in each frequency band
    for k in range(nchan):
        # Extract the envelope in the frequency band
        r = env[k, :]

        # The first (half) windowed segment
        nstart = 0
        smooth[k, 0] = np.sum(r[nstart:nhalf] * halfwindow.conj().transpose()) / halfsum

        # Loop over the remaining full segments, 50% overlap
        for n in range(1, nseg - 1):
            nstart = int(nstart + nhalf)
            nstop = int(nstart + nwin)
            smooth[k, n] = sum(r[nstart:nstop] * window.conj().transpose()) / wsum

        # The last (half) windowed segment
        nstart = nstart + nhalf
        nstop = nstart + nhalf
        smooth[k, nseg - 1] = (
            np.sum(r[nstart:nstop] * window[:nhalf].conj().transpose()) / halfsum
        )

    return smooth


def melcor(x, y, thr, addnoise):
    """
    Function to compute the cross-correlations between the input signal
    time-frequency envelope and the distortion time-frequency envelope. For
    each time interval, the log spectrum is fitted with a set of half-cosine
    basis functions. The spectrum weighted by the basis functions corresponds
    to mel cepstral coefficients computed in the frequency domain. The
    amplitude-normalized cross-covariance between the time-varying basis
    functions for the input and output signals is then computed.

    Arguments:
        x : subsampled input signal envelope in dB SL in each critical band
        y : subsampled distorted output signal envelope
        thr : threshold in dB SPL to include segment in calculation
        addnoise : additive Gaussian noise to ensure 0 cross-corr at low levels

    Returns:
        m1 : average cepstral correlation 2-6, input vs output
        xy : individual cepstral correlations, input vs output

    James M. Kates, 24 October 2006.
    Difference signal removed for cochlear model, 31 January 2007.
    Absolute value added 13 May 2011.
    Changed to loudness criterion for silence threhsold, 28 August 2012.

    Translated from MATLAB to Python by Gerardo Roa Dabike, September 2022.
    """

    # Processing parameters
    nbands = x.shape[0]

    # Mel cepstrum basis functions (mel cepstrum because of auditory bands)
    nbasis = 6  # Number of cepstral coefficients to be used
    freq = np.arange(nbasis)
    k = np.arange(nbands)
    cepm = np.zeros((nbands, nbasis))
    for nb in range(nbasis):
        basis = np.cos(k * float(freq[nb]) * np.pi / float((nbands - 1)))
        cepm[:, nb] = basis / np.linalg.norm(basis)

    # Find the segments that lie sufficiently above the quiescent rate
    xLinear = 10 ** (x / 20)  # Convert envelope dB to linear (specific loudness)
    xsum = np.sum(xLinear, 0) / nbands  # Proportional to loudness in sones
    xsum = 20 * np.log10(xsum)  # Convert back to dB (loudness in phons)
    index = np.where(xsum > thr)[0]  # Identify those segments above threshold
    nsamp = index.shape[0]  # Number of segments above threshold

    # Exit if not enough segments above zero
    m1 = 0
    xy = 0
    if nsamp <= 1:
        print("Function eb.melcor: Signal below threshold, outputs set to 0.")
        return m1, xy

    # Remove the silent intervals
    x = x[:, index]
    y = y[:, index]

    # Add the low-level noise to the envelopes
    x = x + addnoise * np.random.standard_normal(x.shape)
    y = y + addnoise * np.random.standard_normal(y.shape)

    # Compute the mel cepstrum coefficients using only those segments
    # above threshold
    xcep = np.zeros((nbasis, nsamp))  # Input
    ycep = np.zeros((nbasis, nsamp))  # Output
    for n in range(nsamp):
        for k in range(nbasis):
            xcep[k, n] = np.sum(x[:, n] * cepm[:, k])
            ycep[k, n] = np.sum(y[:, n] * cepm[:, k])

    # Remove the average value from the cepstral coefficients. The
    # cross-correlation thus becomes a cross-covariance, and there
    # is no effect of the absolute signal level in dB.
    for k in range(nbasis):
        xcep[k, :] = xcep[k, :] - np.mean(xcep[k, :], axis=0)
        ycep[k, :] = ycep[k, :] - np.mean(ycep[k, :], axis=0)

    # Normalized cross-correlations between the time-varying cepstral coeff
    xy = np.zeros(nbasis)  # Input vs output
    small = 1.0e-30
    for k in range(nbasis):
        xsum = np.sum(xcep[k, :] ** 2)
        ysum = np.sum(ycep[k, :] ** 2)
        if (xsum < small) or (ysum < small):
            xy[k] = 0.0
        else:
            xy[k] = np.abs(np.sum(xcep[k, :] * ycep[k, :]) / np.sqrt(xsum * ysum))

    #
    # % Figure of merit is the average of the cepstral correlations, ignoring
    # % the first (average spectrum level).
    m1 = np.sum(xy[1:nbasis]) / (nbasis - 1)
    return m1, xy


def melcor9(x, y, thr, addnoise, segsize):
    """
    Function to compute the cross-correlations between the input signal
    time-frequency envelope and the distortion time-frequency envelope. For
    each time interval, the log spectrum is fitted with a set of half-cosine
    basis functions. The spectrum weighted by the basis functions corresponds
    to mel cepstral coefficients computed in the frequency domain. The
    amplitude-normalized cross-covariance between the time-varying basis
    functions for the input and output signals is then computed for each of
    the 8 modulation frequencies.

    Arguments:
        x : subsampled input signal envelope in dB SL in each critical band
        y : subsampled distorted output signal envelope
        thr : threshold in dB SPL to include segment in calculation
        addnoise : additive Gaussian noise to ensure 0 cross-corr at low levels
        segsize : segment size in ms used for the envelope LP filter (8 msec)

    Returns:
        CMave : average of the modulation correlations across analysis frequency
            bands and modulation frequency bands, basis functions 2 -6
        CMlow : average over the four lower mod freq bands, 0 - 20 Hz
        CMhigh : average over the four higher mod freq bands, 20 - 125 Hz
        CMmod : vector of cross-correlations by modulation frequency,
            averaged over ananlysis frequency band

    James M. Kates, 24 October 2006.
    Difference signal removed for cochlear model, 31 January 2007.
    Absolute value added 13 May 2011.
    Changed to loudness criterion for silence threshold, 28 August 2012.
    Version using envelope modulation filters, 15 July 2014.
    Modulation frequency vector output added 27 August 2014.

    Translated from MATLAB to Python by Gerardo Roa Dabike, September 2022.
    """

    # Processing parameters
    nbands = x.shape[0]

    # Mel cepstrum basis functions (mel cepstrum because of auditory bands)
    nbasis = 6  # Number of cepstral coefficients to be used
    freq = np.arange(nbasis)
    k = np.arange(nbands)
    cepm = np.zeros((nbands, nbasis))
    for nb in range(nbasis):
        basis = np.cos(k * float(freq[nb]) * np.pi / float((nbands - 1)))
        cepm[:, nb] = basis / np.linalg.norm(basis)

    # Find the segments that lie sufficiently above the quiescent rate
    xLinear = 10 ** (x / 20)  # Convert envelope dB to linear (specific loudness)
    xsum = np.sum(xLinear, 0) / nbands  # Proportional to loudness in sones
    xsum = 20 * np.log10(xsum)  # Convert back to dB (loudness in phons)
    index = np.where(xsum > thr)[0]  # Identify those segments above threshold
    nsamp = index.shape[0]  # Number of segments above threshold

    # Modulation filter bands, segment size is 8 msec
    edge = [4.0, 8.0, 12.5, 20.0, 32.0, 50.0, 80.0]  # 8 bands covering 0 to 125 Hz
    nmod = 1 + len(edge)  # Number of modulation filter bands

    # Exit if not enough segments above zero
    CMave = 0
    CMlow = 0
    CMhigh = 0
    CMmod = np.zeros(nmod)
    if nsamp <= 1:
        print("Function eb.melcor9: Signal below threshold, outputs set to 0.")
        return CMave, CMlow, CMhigh, CMmod

    # Remove the silent intervals
    x = x[:, index]
    y = y[:, index]

    # Add the low-level noise to the envelopes
    x = x + addnoise * np.random.standard_normal(x.shape)
    y = y + addnoise * np.random.standard_normal(y.shape)

    # Compute the mel cepstrum coefficients using only those segments
    # above threshold
    xcep = np.zeros((nbasis, nsamp))  # Input
    ycep = np.zeros((nbasis, nsamp))  # Output
    for n in range(nsamp):
        for k in range(nbasis):
            xcep[k, n] = np.sum(x[:, n] * cepm[:, k])
            ycep[k, n] = np.sum(y[:, n] * cepm[:, k])

    # Remove the average value from the cepstral coefficients. The
    # cross-correlation thus becomes a cross-covariance, and there
    # is no effect of the absolute signal level in dB.
    for k in range(nbasis):
        xcep[k, :] = xcep[k, :] - np.mean(xcep[k, :], axis=0)
        ycep[k, :] = ycep[k, :] - np.mean(ycep[k, :], axis=0)

    # Envelope sampling parameters
    fsub = 1000.0 / (0.5 * segsize)  # Envelope sampling frequency in Hz
    fnyq = 0.5 * fsub  # Envelope Nyquist frequency

    # Design the linear-phase envelope modulation filters
    nfir = np.around(128 * (fnyq / 125))  # Adjust filter length to sampling rate
    nfir = int(2 * np.floor(nfir / 2))  # Force an even filter length
    b = np.zeros((nmod, nfir + 1))

    b[0, :] = firwin(
        nfir + 1, edge[0] / fnyq, window="hann", pass_zero="lowpass"
    )  # LP filter 0-4 Hz
    b[nmod - 1, :] = firwin(
        nfir + 1, edge[nmod - 2] / fnyq, window="hann", pass_zero="highpass"
    )  # HP 80-125 Hz
    for m in range(1, nmod - 1):
        b[m, :] = firwin(
            nfir + 1,
            [edge[m - 1] / fnyq, edge[m] / fnyq],
            window="hann",
            pass_zero="bandpass",
        )  # Bandpass filter

    CM = melcor9_crosscovmatrix(b, nmod, nbasis, nsamp, nfir, xcep, ycep)

    # Average over the  modulation filters and basis functions 2 - 6
    for m in range(nmod):
        for j in range(1, nbasis):
            CMave += CM[m, j]

    CMave = CMave / (nmod * (nbasis - 1))

    # Average over the four lower modulation filters
    for m in range(4):
        for j in range(1, nbasis):
            CMlow += CM[m, j]

    CMlow = CMlow / (4 * (nbasis - 1))

    #  Average over the four upper modulation filters
    for m in range(4, 8):
        for j in range(1, nbasis):
            CMhigh += CM[m, j]

    CMhigh = CMhigh / (4 * (nbasis - 1))

    # Average each modulation frequency over the basis functions
    for m in range(nmod):
        ave = 0
        for j in range(1, nbasis):
            ave += CM[m, j]

        CMmod[m] = ave / (nbasis - 1)

    return CMave, CMlow, CMhigh, CMmod


def melcor9_crosscovmatrix(b, nmod, nbasis, nsamp, nfir, xcep, ycep):
    """Compute the cross-covariance matrix."""
    small = 1.0e-30
    nfir2 = nfir / 2
    # Convolve the input and output envelopes with the modulation filters
    X = np.zeros((nmod, nbasis, nsamp))
    Y = np.zeros((nmod, nbasis, nsamp))
    for m in range(nmod):
        for j in range(nbasis):
            c = convolve(b[m], xcep[j, :], mode="full")
            X[m, j, :] = c[int(nfir2) : int(nfir2 + nsamp)]  # Remove the transients
            c = convolve(b[m], ycep[j, :], mode="full")
            Y[m, j, :] = c[int(nfir2) : int(nfir2 + nsamp)]  # Remove the transients

    # Compute the cross-covariance matrix
    CM = np.zeros((nmod, nbasis))
    for m in range(nmod):
        for j in range(nbasis):
            #  Index j gives the input reference band
            xj = X[m, j]  # Input freq band j, modulation freq m
            xj = xj - np.mean(xj)
            xsum = np.sum(xj**2)

            # Processed signal band
            yj = Y[m, j]  # Input freq band j, modulation freq m
            yj = yj - np.mean(yj)
            ysum = np.sum(yj**2)

            # Cross-correlate the reference and processed signals
            if (xsum < small) or (ysum < small):
                CM[m, j] = 0
            else:
                CM[m, j] = np.abs(np.sum(xj * yj)) / np.sqrt(xsum * ysum)
    return CM


def spect_diff(xSL, ySL):
    """
    Function to compute changes in the long-term spectrum and spectral slope.
    The metric is based on the spectral distortion metric of Moore and Tan
    (JAES, Vol 52, pp 900-914). The log envelopes in dB SL are converted to
    linear to approximate specific loudness. The outputs are the sum of the
    absolute differences, the standard deviation of the differences, and the
    maximum absolute difference. The same three outputs are provided for the
    normalized spectral difference and for the slope. The output is
    calibrated so that a processed signal having 0 amplitude produces a
    value of 1 for the spectrum difference.

    Abs diff: weight all deviations uniformly
    Std diff: weight larger deviations more than smaller deviations
    Max diff: only weight the largest deviation

    Arguments:
        xSL : reference signal spectrum in dB SL
        ySL : degraded signal spectrum in dB SL

    Returns:
        dloud (np.array) : [sum abs diff, std dev diff, max diff] spectra
        dnorm (np.array) : [sum abs diff, std dev diff, max diff] norm spectra
        dslope (np.array) : [sum abs diff, std dev diff, max diff] slope

    James M. Kates, 28 June 2012.

    Translated from MATLAB to Python by Gerardo Roa Dabike, September 2022.
    """

    # Convert the dB SL to linear magnitude values. Because of the auditory
    # filter bank, the OHC compression, and auditory threshold, the linear
    # values are closely related to specific loudness.
    nbands = xSL.shape[0]
    x = 10 ** (xSL / 20)
    y = 10 ** (ySL / 20)

    # Normalize the level of the reference and degraded signals to have the
    # same loudness. Thus overall level is ignored while differences in
    # spectral shape are measured.
    xsum = np.sum(x)
    x /= xsum  # Loudness sum = 1 (arbitrary amplitude, proportional to sones)
    ysum = np.sum(y)
    y /= ysum

    # Compute the spectrum difference
    dloud = np.zeros(3)
    d = x - y  # Difference in specific loudness in each band
    dloud[0] = np.sum(np.abs(d))
    dloud[1] = nbands * np.std(d)  # Biased std: second moment
    dloud[2] = np.max(np.abs(d))

    # Compute the normalized spectrum difference
    dnorm = np.zeros(3)
    d = (x - y) / (x + y)  # Relative difference in specific loudness
    dnorm[0] = np.sum(np.abs(d))
    dnorm[1] = nbands * np.std(d)
    dnorm[2] = np.max(np.abs(d))

    # Compute the slope difference
    dslope = np.zeros(3)
    dx = x[1:nbands] - x[0 : nbands - 1]
    dy = y[1:nbands] - y[0 : nbands - 1]
    d = dx - dy  # Slope difference
    dslope[0] = np.sum(np.abs(d))
    dslope[1] = nbands * np.std(d)
    dslope[2] = np.max(np.abs(d))

    return dloud, dnorm, dslope


def bm_covary(xBM, yBM, segsize, fsamp):
    """
    Function to compute the cross-covariance (normalized cross-correlation)
    between the reference and processed signals in each auditory band. The
    signals are divided into segments having 50% overlap.

    Arguments:
        xBM : BM movement, reference signal
        yBM : BM movement, processed signal
        segsize : signal segment size, msec
        fsamp : sampling rate in Hz

    Returns:
        sigcov (np.array) : [nchan,nseg] of cross-covariance values
        sigMSx (np.array) : [nchan,nseg] of MS input signal energy values
        sigMSy (np.array) : [nchan,nseg] of MS processed signal energy values

    James M. Kates, 28 August 2012.
    Output amplitude adjustment added, 30 october 2012.

    Translated from MATLAB to Python by Gerardo Roa Dabike, September 2022.
    """

    # Initialize parameters
    small = 1.0e-30

    # Lag for computing the cross-covariance
    lagsize = 1.0  # Lag (+/-) in msec
    maxlag = np.around(lagsize * (0.001 * fsamp))  # Lag in samples

    # Compute the segment window
    nwin = int(np.around(segsize * (0.001 * fsamp)))  # Segment size in samples
    test = nwin - 2 * np.floor(nwin / 2)  # 0=even, 1=odd
    if test > 0:
        nwin = nwin + 1  # Force window length to be even

    window = np.hanning(nwin).conj().transpose()  # Raised cosine von Hann window
    wincorr = correlate(window, window, "full")  # Window autocorrelation, inverted
    wincorr = wincorr[int(len(window) - 1 - maxlag) : int(maxlag + len(window))]
    wincorr = 1 / wincorr
    winsum2 = 1.0 / np.sum(window**2)  # Window power, inverted

    # The first segment has a half window
    nhalf = int(nwin / 2)
    halfwindow = window[nhalf:nwin]
    halfcorr = correlate(halfwindow, halfwindow, "full")
    halfcorr = halfcorr[
        int(len(halfwindow) - 1 - maxlag) : int(maxlag + len(halfwindow))
    ]
    halfcorr = 1 / halfcorr
    halfsum2 = 1.0 / np.sum(halfwindow**2)  # MS sum normalization, first segment

    # Number of segments
    nchan = xBM.shape[0]
    npts = xBM.shape[1]
    nseg = int(1 + np.floor(npts / nwin) + np.floor((npts - nwin / 2) / nwin))
    sigMSx = np.zeros((nchan, nseg))
    sigMSy = np.zeros((nchan, nseg))
    sigcov = np.zeros((nchan, nseg))

    # Loop to compute the signal mean-squared level in each band for each
    # segment and to compute the cross-corvariances.
    for k in range(nchan):
        # Extract the BM motion in the frequency band
        x = xBM[k, :]
        y = yBM[k, :]

        # The first (half) windowed segment
        nstart = 0
        segx = x[nstart:nhalf] * halfwindow  # Window the reference
        segy = y[nstart:nhalf] * halfwindow  # Window the processed signal
        segx = segx - np.mean(segx)  # Make 0-mean
        segy = segy - np.mean(segy)
        MSx = np.sum(segx**2) * halfsum2  # Normalize signal MS value by the window
        MSy = np.sum(segy**2) * halfsum2
        c = correlate(segx, segy, "full")
        c = c[int(len(segx) - 1 - maxlag) : int(maxlag + len(segx))]
        Mxy = np.max(np.abs(c * halfcorr))  # Unbiased cross-correlation
        if (MSx > small) and (MSy > small):
            sigcov[k, 0] = Mxy / np.sqrt(MSx * MSy)  # Normalized cross-covariance
        else:
            sigcov[k, 0] = 0.0

        sigMSx[k, 0] = MSx  # Save the reference MS level
        sigMSy[k, 0] = MSy

        # Loop over the remaining full segments, 50% overlap
        for n in range(1, nseg - 1):
            nstart = nstart + nhalf
            nstop = nstart + nwin
            segx = x[nstart:nstop] * window  # Window the reference
            segy = y[nstart:nstop] * window  # Window the processed signal
            segx = segx - np.mean(segx)  # Make 0-mean
            segy = segy - np.mean(segy)
            MSx = np.sum(segx**2) * winsum2  # Normalize signal MS value by the window
            MSy = np.sum(segy**2) * winsum2
            c = correlate(segx, segy, "full")
            c = c[int(len(segx) - 1 - maxlag) : int(maxlag + len(segx))]
            Mxy = np.max(np.abs(c * wincorr))  # Unbiased cross-corr
            if (MSx > small) and (MSy > small):
                sigcov[k, n] = Mxy / np.sqrt(MSx * MSy)  # Normalized cross-covariance
            else:
                sigcov[k, n] = 0.0

            sigMSx[k, n] = MSx  # Save the reference MS level
            sigMSy[k, n] = MSy  # Save the reference MS level

        # The last (half) windowed segment
        nstart = nstart + nhalf
        nstop = nstart + nhalf
        segx = x[nstart:nstop] * window[0:nhalf]  # Window the reference
        segy = y[nstart:nstop] * window[0:nhalf]  # Window the processed signal
        segx = segx - np.mean(segx)  # Make 0-mean
        segy = segy - np.mean(segy)
        MSx = np.sum(segx**2) * halfsum2  # Normalize signal MS value by the window
        MSy = np.sum(segy**2) * halfsum2

        c = np.correlate(segx, segy, "full")
        c = c[int(len(segx) - 1 - maxlag) : int(maxlag + len(segx))]

        Mxy = np.max(np.abs(c * halfcorr))  # Unbiased cross-correlation
        if (MSx > small) and (MSy > small):
            sigcov[k, nseg - 1] = Mxy / np.sqrt(
                MSx * MSy
            )  # Normalized cross-covariance
        else:
            sigcov[k, nseg - 1] = 0.0

        sigMSx[k, nseg - 1] = MSx  # Save the reference MS level
        sigMSy[k, nseg - 1] = MSy  # Save the reference MS level

    # Limit the cross-covariance to lie between 0 and 1
    sigcov = np.clip(sigcov, 0, 1)

    # Adjust the BM magnitude to correspond to the envelope in dB SL
    sigMSx = 2.0 * sigMSx
    sigMSy = 2.0 * sigMSy

    return sigcov, sigMSx, sigMSy


def ave_covary2(sigcov, sigMSx, thr):
    """
    Function to compute the average cross-covariance between the reference
    and processed signals in each auditory band. The silent time-frequency
    tiles are removed from consideration. The cross-covariance is computed
    for each segment in each frequency band. The values are weighted by 1
    for inclusion or 0 if the tile is below threshold. The sum of the
    covariance values across time and frequency are then divided by the total
    number of tiles above thresold. The calculation is a modification of
    Tan et al. (JAES 2004). The cross-covariance is also output with a
    frequency weighting that reflects the loss of IHC synchronization at high
    frequencies (Johnson, JASA 1980).

    Arguments:
        sigcov (np.array) : [nchan,nseg] of cross-covariance values
        sigMSx (np.array) : [nchan,nseg] of reference signal MS values
        thr : threshold in dB SL to include segment ave over freq in average

    Returns:
        avecov : cross-covariance in segments averaged over time and frequency
        syncov : cross-coraviance array, 6 different weightings for loss of
              IHC synchronization at high frequencies:
              LP Filter Order     Cutoff Freq, kHz
                1              1.5
                3              2.0
                5              2.5, 3.0, 3.5, 4.0

    James M. Kates, 28 August 2012.
    Adjusted for BM vibration in dB SL, 30 October 2012.
    Threshold for including time-freq tile modified, 30 January 2013.
    Version for different sync loss, 15 February 2013.

    Translated from MATLAB to Python by Gerardo Roa Dabike, September 2022.
    """

    # Array dimensions
    nchan = sigcov.shape[0]

    # Initialize the LP filter for loss of IHC synchronization
    cfreq = CenterFreq(nchan)  # Center frequencies in Hz on an ERB scale
    p = np.array([1, 3, 5, 5, 5, 5])  # LP filter order
    fcut = 1000 * np.array([1.5, 2.0, 2.5, 3.0, 3.5, 4.0])  # Cutoff frequencies in Hz
    fsync = np.zeros((6, nchan))  # Array of filter freq resp vs band center freq
    for n in range(6):
        fc2p = fcut[n] ** (2 * p[n])
        freq2p = cfreq ** (2 * p[n])
        fsync[n, :] = np.sqrt(fc2p / (fc2p + freq2p))

    # Find the segments that lie sufficiently above the threshold.
    sigRMS = np.sqrt(sigMSx)  # Convert squared amplitude to dB envelope
    sigLinear = 10 ** (sigRMS / 20)  # Linear amplitude (specific loudness)
    xsum = np.sum(sigLinear, 0) / nchan  # Intensity averaged over frequency bands
    xsum = 20 * np.log10(xsum)  # Convert back to dB (loudness in phons)
    index = np.argwhere(xsum > thr).T  # Identify those segments above threshold
    if index.size != 1:
        index = index.squeeze()
    nseg = index.shape[0]  # Number of segments above threshold

    # Exit if not enough segments above zero
    if nseg <= 1:
        print("Function eb.AveCovary: Ave signal below threshold, outputs set to 0.")
        avecov = 0
        # syncov = 0
        syncov = [0] * 6
        return avecov, syncov

    # Remove the silent segments
    sigcov = sigcov[:, index]
    sigRMS = sigRMS[:, index]

    # Compute the time-frequency weights. The weight=1 if a segment in a
    # frequency band is above threshold, and weight=0 if below threshold.
    weight = np.zeros((nchan, nseg))  # No IHC synchronization roll-off
    wsync1 = np.zeros((nchan, nseg))  # Loss of IHC synchronization at high frequencies
    wsync2 = np.zeros((nchan, nseg))
    wsync3 = np.zeros((nchan, nseg))
    wsync4 = np.zeros((nchan, nseg))
    wsync5 = np.zeros((nchan, nseg))
    wsync6 = np.zeros((nchan, nseg))
    for k in range(nchan):
        for n in range(nseg):
            if sigRMS[k, n] > thr:  # Thresh in dB SL for including time-freq tile
                weight[k, n] = 1
                wsync1[k, n] = fsync[0, k]
                wsync2[k, n] = fsync[1, k]
                wsync3[k, n] = fsync[2, k]
                wsync4[k, n] = fsync[3, k]
                wsync5[k, n] = fsync[4, k]
                wsync6[k, n] = fsync[5, k]

    # Sum the weighted covariance values
    csum = np.sum(np.sum(weight * sigcov))  # Sum of weighted time-freq tiles
    wsum = np.sum(np.sum(weight))  # Total number of tiles above thresold
    fsum = np.zeros(6)
    ssum = np.zeros(6)
    fsum[0] = np.sum(np.sum(wsync1 * sigcov))  # Sum of weighted time-freq tiles
    ssum[0] = np.sum(np.sum(wsync1))  # Total number of tiles above thresold
    fsum[1] = np.sum(np.sum(wsync2 * sigcov))  # Sum of weighted time-freq tiles
    ssum[1] = np.sum(np.sum(wsync2))  # Total number of tiles above thresold
    fsum[2] = np.sum(np.sum(wsync3 * sigcov))  # Sum of weighted time-freq tiles
    ssum[2] = np.sum(np.sum(wsync3))  # Total number of tiles above thresold
    fsum[3] = np.sum(np.sum(wsync4 * sigcov))  # Sum of weighted time-freq tiles
    ssum[3] = np.sum(np.sum(wsync4))  # Total number of tiles above thresold
    fsum[4] = np.sum(np.sum(wsync5 * sigcov))  # Sum of weighted time-freq tiles
    ssum[4] = np.sum(np.sum(wsync5))  # Total number of tiles above thresold
    fsum[5] = np.sum(np.sum(wsync6 * sigcov))  # Sum of weighted time-freq tiles
    ssum[5] = np.sum(np.sum(wsync6))  # Total number of tiles above thresold

    # Exit if not enough segments above zero
    if wsum < 1:
        avecov = 0
        print("Function eb.AveCovary: Signal tiles below threshold, outputs set to 0.")
    else:
        avecov = csum / wsum
    syncov = fsum / ssum

    return avecov, syncov
